/* Svgalib Web Browser by Clocksoft */
/* (c) 1999 Clocksoft */

#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <signal.h>

#include "gvars.h"
#include "scroller.h"
#include "line_breaker.h"
#include "block.h"

#undef DEBUG_LINE_BREAKER
#define ISSPACE(c) ((c)==32||(c)==9)
#define FONT_SPACE ' '

int find_break_length(unsigned char *txt,int length)
/* Finds the length of the text to remain on the last line. */
/* There is no space at the end. */
{int pos;
 
 for (pos=length-1;pos>=0;pos--)
 {if (ISSPACE(txt[pos])) return pos;
 }
 return length; /* The line doesn't contain spaces and is as long as it needs */
 /* to be broken. All the characters will remain on the last line, no hyphen */
 /* will be created (i hope this would be very rare case). */
}

int recompute_line_width(unsigned char *txt,int length)
{int ctr;
 unsigned char *t;
 int total;

 for (ctr=length,t=txt,total=0;ctr;ctr--,t++)
  total+=font_widths[*t];

 return total;
}

/* Doesn't destroy the input. xallocs new space. */
bd
wrap_chars(bd in)
{
 bd out;
 if (in.number<256)
 {
  out=new_block(in.number+2);
  memcpy(out.ptr+2,in.ptr,in.number);
  out.ptr[1]=in.number;
  *out.ptr=0; /* Short block of chars */
 }
 else
 {
  unsigned long a=in.number;
  out=new_block(in.number+5);
  memcpy(out.ptr+5,in.ptr,in.number);
  *out.ptr=1; /* Long block of chars */
  out.ptr[1]=a;
  out.ptr[2]=a>>8;
  out.ptr[3]=a>>16;
  out.ptr[4]=a>>24;
 }
 return out;
}

bd
wrap_bg(int r, int g, int b)
{
 bd out=new_block(4);
 *out.ptr=2;
 out.ptr[1]=r;
 out.ptr[2]=g;
 out.ptr[3]=b;
 return out;
}

bd
wrap_fg(int r, int g, int b)
{
 bd out=new_block(4);
 *out.ptr=3;
 out.ptr[1]=r;
 out.ptr[2]=g;
 out.ptr[3]=b;
 return out;
}

bd
concat(bd in1, bd in2)
{
 bd out=new_block(in1.number+in2.number);
 memcpy(out.ptr,in1.ptr,in1.number);
 memcpy(out.ptr+in1.number,in2.ptr,in2.number);
 return out;
}

int line_breaker(int command,void *params)
{static int line_valid=0;
 static unsigned char *line=NULL;
 static int line_bts,line_width;
 static int line_pos;
 static int state=LBI_SPACE; /* At the beginning, spaces will be removed */
 struct block_desc bde;
 int ch_width;
 unsigned char c;
 /* line_valid tells whether the line is malloced */
 /* line_bts tells the arg for malloc for line and is also equal
    the number of pixels line can max. contain */
 /* line_width tells the total width of liner content in pixels. */
 /* line_pos tells number of chars contained in line. */

#ifdef DEBUG_LINE_BREAKER
 printf("I got into line breaker with command=%d.\n",command);
#endif
 switch(command)
 {
  case LB_NEW_SCREEN:
  line_bts=xsize;
  free(line);
  line=NULL;
  line_valid=0;
  if (scroller(SC_NEW_SCREEN,params)) /* params is pal_mode */
  {
   fprintf(stderr,"Unable to scroller(SC_NEW_SCREEN).\n");
   return 1;
  }
  break;
 
  case LB_NEW_FONT:
  free(line);
  line=NULL;
  line_valid=0;
  if (scroller(SC_NEW_FONT,params))
  {
   fprintf(stderr,"Unable to scroller(SC_NEW_FONT).\n");
   return 1;
  }
  break;
 
  case LB_TYPESET_BLOCK:
  bde=*((struct block_desc *)params);
  #ifdef DEBUG_LINE_BREAKER
   printf("I got into LB_TYPESET_BLOCK, length=%d.\n",bde.number);
  #endif
  if (!bde.number) return 0; /* No chars in block */ 
  if (!line_valid)
  {if (!(line=(unsigned char *)malloc(line_bts)))
   {fprintf(stderr,"LB: Can't alloc line buffer.\n");return 1;}
#ifdef DEBUG_LINE_BREAKER
   printf("Allocated line of length=%d.\n",line_bts);
#endif
   line_pos=0;
   line_valid=1;
   line_width=0;
   state=LBI_SPACE; /* New line, spaces will be removed. */
  }
  do 
  {c=ISSPACE(*bde.ptr)?FONT_SPACE:*bde.ptr;
   ch_width=font_widths[c];
   if (!ch_width||(c==FONT_SPACE&&state==LBI_SPACE)) /* Char with empty image */
   /* or second space */
   {bde.ptr++;
    bde.number--;
    continue; /* Take next char. */
   }
   if (line_width+ch_width>line_bts)
   {/* We are gonna break */
    int break_len;
    #ifdef DEBUG_LINE_BREAKER
     printf("Breaking line, line_pos=%d,",line_pos);
    #endif
    if (c==FONT_SPACE)
     break_len=line_pos; /* If space caused break, all the content will */
                         /* remain on the original line. */
    else
     break_len=find_break_length(line,line_pos);
    #ifdef DEBUG_LINE_BREAKER
     printf("break_len=%d.\n",break_len);
    #endif
    /* Finds length of line from the beginning which is to be shipped out
       and the rest is to be broken onto the next line. We have to watch out:
       if the rest starts with space, we have to remove it. */
    /* last_space won't be more than line_buffer_pos */
    {bd sc_block;
     bd block1,block2,block3,block4,block5;
     
     sc_block.ptr=line;
     sc_block.number=break_len;
     if (image_mode)
     {
      if (scroller(SC_ADD_LINE,&sc_block)) goto err1;
     }
     else
     {
      block1=wrap_chars(sc_block);
      block2=wrap_bg(0,0,0);
      block3=wrap_fg(255,255,255);
      block4=concat(block2,block3);
      free(block2.ptr);
      free(block3.ptr);
      block5=concat(block4,block1);
      free(block4.ptr);
      free(block1.ptr);
      if (scroller(SC_ADD_LINE,&block5)) return 1;
      free(block5.ptr);
     }
     /* Line shipped out. */
    }
    /* Rest of line: line[break_len]...line[line_pos] */
    if (break_len<line_pos&&line[break_len]==FONT_SPACE) break_len++;
    /* Now break len tells the begin of content of next line */
    memmove(line,line+break_len,line_pos-break_len);
    line_pos-=break_len;
    if (c==FONT_SPACE)
    {state=LBI_SPACE;
     /* If a space caused break throw it out and go into space state */
    }
    else
    {state=LBI_NORMAL;
     line[line_pos]=c; /* Write the char which caused the break */
     line_pos++;
    }
    bde.ptr++;
    bde.number--;
    line_width=recompute_line_width(line,line_pos);
   }
   else
   {
    /* No break */
    /* Now the branch where there is enough space for the new char */
    /* And the char is sure to be written (no spc after spc) */
    /* And it's at least one pixel wide. */
    line[line_pos++]=c;
    line_width+=font_widths[c];
    if (c==FONT_SPACE)
     state=LBI_SPACE;
    else
     state=LBI_NORMAL; /* Update space */
    bde.ptr++;bde.number--;
  }
  }
  while(bde.number);
  return 0;
  break;
  err1:
  fprintf(stderr,"line_breaker: Can't typeset block.\n");
  return 1;
  
  case LB_FLUSH_LINE:
  {struct block_desc sc_block,block1,block2,block3,block4,block5;
   sc_block.ptr=line;
   sc_block.number=line_pos;
   
   if (image_mode)
   {
    if (scroller(SC_ADD_LINE,&sc_block)) goto err2;
   }
   else
   {
    block1=wrap_chars(sc_block);
    block2=wrap_bg(0,0,0);
    block3=wrap_fg(255,255,255);
    block4=concat(block2,block3);
    free(block2.ptr);
    free(block3.ptr);
    block5=concat(block4,block1);
    free(block4.ptr);
    free(block1.ptr);
    if (scroller(SC_ADD_LINE,&block5)) return 1;
    free(block5.ptr);
   }
   /* Rest of line shipped out. */
  }
  line_width=0;line_pos=0;state=LBI_SPACE; /* Update state */
  break;
  err2:
  fprintf(stderr,"line_breaker: Can't flush line.\n");
  return 1;

 }
 return 0;
}

