#include <unistd.h>
#include <stdio.h>
#include <string.h>

#define PT_EPSON 0
#define PT_HP_QUIETJET 1
int printer_type=PT_EPSON;
static int handle;
int vquality, hquality;

#define PXSIZE 1920

#define max(x,y) ((x)>(y)?(x):(y))
#define min(x,y) ((x)<(y)?(x):(y))

unsigned strip[PXSIZE]; /* Lower byte first pass (even lines), higher second
                            * (odd lines). Starting counting lines from 0.
                            */
unsigned strip_masks[24]=
{
 0x80,0x8000,0x800000,
 0x40,0x4000,0x400000,
 0x20,0x2000,0x200000,
 0x10,0x1000,0x100000,
 0x08,0x0800,0x080000,
 0x04,0x0400,0x040000,
 0x02,0x0200,0x020000,
 0x01,0x0100,0x010000
};

int strip_mask_index;

extern volatile int lpr_runs, lpr_pid;

void
start_graphics(int length, int hquality)
{
 unsigned char c[3]={'K','L','Z'};
 switch(hquality)
 {
  case 3: break;
  case 2: length=(length*2+2)/3; break;
  case 1: length=(length+2)/3; break;
 }
 {
  unsigned char t[]={27,c[hquality-1],length,length>>8}; 
  write(handle,&t,sizeof(t));
 }
}

void
short_skip(void)
{
 unsigned char t[]={27,'J',1,13};
 write(handle,&t,sizeof(t));
}

void
long_skip(void)
{
 unsigned char t[]={27,'J',25-vquality,13};
 write(handle,&t,sizeof(t));
}

void
print_strip(void)
{
 int i,x;
 unsigned char c;
 int length;
 unsigned mask;
 int shift;

 if (hquality<1) hquality=1;
 if (hquality>3) hquality=3;
 if (vquality<1) vquality=1;
 if (vquality>3) vquality=3;

 switch(vquality)
 {
  case 3: break;
  case 2:
  for (x=0;x<PXSIZE;x++)
  {
   strip[x]|=(strip[x]>>8)&0xff00;
  }
  break;
  case 1:
  for (x=0;x<PXSIZE;x++)
  {
   strip[x]|=(strip[x]>>16)|strip[x]>>8;
  }
  break;
 }
 switch(hquality)
 {
  case 3: break;
  case 2:
  for (x=1;x<PXSIZE-1;x+=2)
   strip[x]|=strip[x+1];
  break;
  case 1:
  for (x=0;x<PXSIZE-1;x+=3)
   strip[x]|=strip[x+1];
  for (x=0;x<PXSIZE-2;x+=3)
   strip[x]|=strip[x+2];
  break;
 }
 for (shift=0;shift<8*vquality;shift+=8)
 {
  mask=0xff<<shift;
  length=PXSIZE;
  while(length&&!(strip[length-1]&mask))length--;
  if (length)
  {
   start_graphics(length,hquality);
   for (i=0;i<length;i++)
   {
    if (i%3<hquality)
    {
     c=strip[i]>>shift;
     write(handle,&c,1);
    }
   }
  }
  if (shift==8*(vquality-1))
  {
   long_skip();
  }
  else
  {
   short_skip();
  }
 }
}

/* zero byte=white, nonzero byte=black */
void
add_printer_line(unsigned char *bytes, int length)
{
 int a;
 unsigned strip_mask=strip_masks[strip_mask_index];
 for (a=0;a<min(PXSIZE,length);a++)
 {
  if (bytes[a]) strip[a]|=strip_mask;
 }
 strip_mask_index++;
 if (strip_mask_index>=24)
 {
  strip_mask_index=0;
  print_strip();
  bzero(strip,sizeof(strip));
 }
}

void
write_header(void)
{
 unsigned char str[]=
 {
  27,64, /* Perform reset */
  /* 27,'O',*/  /* Turn the perforation skip off */
  27,'A',8, /* Line spacing to 8 dot rows */
  0x1b,0x55,0, /* Bidirectional */
  0xd /* Carriage return */
 };

 write(handle,str,sizeof(str));
}

/* Returns 1 error, 0 OK
   If <0, error, otherwise OK. */
int
open_printer(void)
{
 int fds[2]; /* 0 for reading, 1 for writing */
 
 if (pipe(fds)) 
 {
  fprintf(stderr,"printer.c: unable to get a pipe to run \"lpr\".\n");
  return 1;
 }
 lpr_pid=fork();
 if (lpr_pid)
 {
  /* I am the parent */
  lpr_runs=1;
  close(fds[0]);
  handle=fds[1];
  write_header();
  bzero(strip,sizeof(strip));
  strip_mask_index=0;
  return 0;
 }
 else
 {
  /* I am the child -- reading from fds[0] */
  close(fds[1]);
  if (-1==dup2(fds[0],0))
  {
   fprintf(stderr,"printer.c: Unable to dup2. Something is strange!\n");
   return -1;
  }
  close(fds[0]);
  execlp("lpr","lpr",NULL);
  fprintf(stderr,"printer.c -- child: Unable to execlp \"lpr\"");
  _exit(-1);
 }
}

/* Call after you flushed all the pixels into the printer */
void
close_printer(void)
{
 if (strip_mask_index)
 {
  print_strip();
 }
 close(handle);
}
