/* Svgalib Web Browser by Clocksoft */
/* (c) 1999 Clocksoft -- Karel Kulhavy, clock@atrey.karlin.mff.cuni.cz */

#include <vga.h>
#include <vgagl.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

#include "gvars.h"
#include "scroller.h"
#include "printer.h"
#include "block.h"

//#define DG_SCROLLER

#define max(x,y) ((x)>(y)?(x):(y))
#define min(x,y) ((x)<(y)?(x):(y))

/******************* Global variables for s-croller ***************************/

/* The number of 64k-page (only in non-linear graphics modes) which is
 * currently selected on the video card.
 */
static int active_page;
static unsigned char * my_graph_mem; /* Screen */

/* This switches off drawing of black microline tails in write_line_to_screen.
 * Used for speeding up initial drawing of images.
 */
static int global_no_tails=0;

/* The colours of the text. Change as the interpreter goes through the text. */
static unsigned char fgcolor_r,fgcolor_g,fgcolor_b;
static unsigned char bgcolor_r,bgcolor_g,bgcolor_b;
static unsigned char slfgcolor_r=255,slfgcolor_g=255,slfgcolor_b=255;
static unsigned char slbgcolor_r=0,slbgcolor_g=0,slbgcolor_b=0;

/* Number of lines dedicated for displaying the document. There is one more
 * line which is used for displaying status etc. In image mode,
 * The number of scanlines dedicated for the image. top_lines must be at least
 * one. (Otherwise, when font_height>=ysize, there would be no space for
 * viewing the image)
 */
static int top_lines;

/* How many lines to skip when page up/dn is pressed. */
static int page_skip; 

/* Status line, always one whole line high
 * Organization: xsize bytes allocated.
 * Always contains text data.
 */
static unsigned char *sl=NULL; 

/* Number of valid characters in sl */
static int sl_length;

/* sc_b_line tells the number of the top line of the screen */
/* sc_b_line can have any value >=0 */
static int sc_b_line;

/* lines_pos tells number of valid lines in "lines" data structure. */
static int lines_pos=0;

/* Defines mode of scroller's operation
0 text mode
1 image mode
*/
unsigned char image_mode;

/* xoff: For shifting images right-left only. xoff=4 means the pixels
         0,1,2 and 3 of each line are not visible.
   max_width: the width of the widest line in the screen.
             Defined only in image mode, can contain garbage in text mode.
   x_step: a step which is performed on SC_RIGHT and SC_LEFT. Not defined in text
          mode.
*/
int x_off=0,max_width=0,x_step=0;

/* Storage for lines itself */
static struct block_desc *lines=NULL;

/* Number of lines allocated in lines. Not all of them may be initialized!
   The "lines_pos" tell the number of initialized.
*/
static int lines_allocd=0;

/* Actual vga mode definition */
int vga_linewidth; /* Prepared out from vga_getmodeinfo */
int xsize,ysize; /* Prepared out from vga_getmodeinfo */
int vga_bytes; /* Prepared out from vga_getmodeinfo */
int vga_colors; /* prepared out from vga_getmodeinfo */
int vga_linear; /* 1 linear mode, 0 nonlinear mode (paged) */
int vga_misordered; /* The same meaning as misordered in vga_getmodeinfo */ 

/* Dependence of memory layout on vga_bytes and vga_colors:
   -------------------------------------------------------
 
 vga_bytes        
 /--------
 |
 V
 
 1: organization: 76543210 (vga_colors=256)
                  RRRGGGBB
 2: organization: 5432109876543210 (vga_colors=65536)
                   RRRRRGGGGGBBBBB 

                  5432109876543210 (vga_colors=32768)
                  RRRRRGGGGGGBBBBB

 3: organization: 321098765432109876543210 (vga_colors=16777216,
    vga_misordered=0)
                  RRRRRRRRGGGGGGGGBBBBBBBB

 3: organization: 321098765432109876543210 (vga_colors=16777216,
    vga_misordered=1)
                  BBBBBBBBGGGGGGGGRRRRRRRR

 4: organization: 10987654321098765432109876543210 (vga_colors=16777216,
 vga_misordered=0)
                          RRRRRRRRGGGGGGGGBBBBBBBB

 4: organization: 10987654321098765432109876543210 (vga_colors=16777216,
 vga_misordered=1)
                  RRRRRRRRGGGGGGGGBBBBBBBB

*/

int *dither_lut_r;
int *dither_lut_g;
int *dither_lut_b;
int dither_n_r,dither_n_g,dither_n_b;
int dither_e_r,dither_e_g,dither_e_b;

unsigned char gamma_lut_r[256];
unsigned char gamma_lut_g[256];
unsigned char gamma_lut_b[256];
/* : has always 256 entries, for color channel linear values 0...255. For
     each linear value, there is the index of voltage that approximates best
     the correponding voltage to the input value. It means, human's perception
     error is minimized. */

double monitor_gamma_r=2.5,
       monitor_gamma_g=2.5,
       monitor_gamma_b=2.5,
       required_gamma=1,
       monitor_mul_r=1,
       monitor_mul_g=1,
       monitor_mul_b=1,
       monitor_shift_r=0,
       monitor_shift_g=0,
       monitor_shift_b=0;

/* If this flag is zero, background_transcpy and emit_bakground_stream generate
 * background. Otherwise, they generate foreground. */
static int bgtrans_flag;
static int filler_start_scanline,filler_end_scanline;
static const int box_horiz_thickness=2;

static struct char_desc *broken_down_line;
int broken_down_line_length;

/* Uses global variables bgcolor_? which are changed by some
   escape sequences. Performs also the gamma-correction.
   Doesn't take into account any VGA pages.
   If the flag is 1, performs foreground instead of background.
*/
void
background_transcpy(unsigned char *dest, int length)
{
 int o;
 int r,g,b;
 
 
 if (bgtrans_flag)
 {
  r=gamma_lut_r[fgcolor_r];
  g=gamma_lut_g[fgcolor_g];
  b=gamma_lut_b[fgcolor_b];
 }
 else
 {
  r=gamma_lut_r[bgcolor_r];
  g=gamma_lut_g[bgcolor_g];
  b=gamma_lut_b[bgcolor_b];
 }

 switch(vga_colors)
 {
  case 256: o=(r<<5)|(g<<2)|b; break;
  case 32768: o=(r<<10)|(g<<5)|b; break;
  case 65536: o=(r<<11)|(g<<5)|b; break;
  default: /* 16777216 */
  if (vga_bytes==3)
   if (vga_misordered)
    o=(b<<16)|(g<<8)|r;
   else
    o=(r<<16)|(g<<8)|r;
  else
   if (vga_misordered)
    o=(b<<24)|(g<<16)|(b<<8);
   else
    o=(b<<16)|(g<<8)|r;
  break;
 }
 switch (vga_bytes)
 {
  case 1: 
  memset(dest,o,length);
  break;
  
  case 2:
  for (;length;length--)
  {
   *dest=o;
   dest[1]=o>>8;
   dest+=2;
  }
  break;

  case 3:
  for (;length;length--)
  {
   *dest=o;
   dest[1]=o>>8;
   dest[2]=o>>16;
   dest+=3;
  }
  break;

  case 4:
  for (;length;length--)
  {
   *dest=o;
   dest[1]=o>>8;
   dest[2]=o>>16;
   dest[3]=o>>24;
   dest+=4;
  }
  break;
 }
}

/* input is one byte per pixel: 0=background, 255=foreground.
   uses global variables bgcolor_? and fgcolor_? which are changed by some
   escape sequences. Performs also the gamma-correction.
   Doesn't take into account any VGA pages.
   Uses gamma_lut_?[0...255] to obtain the index of color channel to
   bit-cut into the esulting code.
*/
void
text_transcpy(unsigned char *dest, const unsigned char *from, int length)
{
 int w,w1;
 int o;
 int r,g,b;
 
 #define COMPUTE_INDICES w=*from++; w1=255-w;\
  r=(fgcolor_r*w+bgcolor_r*w1);\
  r=(r+1+(r>>8))>>8;\
  r=gamma_lut_r[r];\
  g=(fgcolor_g*w+bgcolor_g*w1);\
  g=(g+1+(g>>8))>>8;\
  g=gamma_lut_g[g];\
  b=(fgcolor_b*w+bgcolor_b*w1);\
  b=(b+1+(b>>8))>>8;\
  b=gamma_lut_b[b];
 /* :The strange bit operations are instead of division by 255. */
 
 switch(vga_colors)
 {
  case 256:
  for (;length;length--)
  {
   COMPUTE_INDICES
   *dest++=(r<<5)|(g<<2)|b;
  }
  break;
  
  case 32768:
  for (;length;length--)
  {
   COMPUTE_INDICES
   o=r<<5;
   o|=g;
   o<<=5;
   o|=b;
   *dest=o;
   dest[1]=o>>8;
   dest+=2;
  }
  break;
  
  case 65536:
  for (;length;length--)
  {
   COMPUTE_INDICES
   o=r<<6;
   o|=g;
   o<<=5;
   o|=b;
   *dest=o;
   dest[1]=o>>8;
   dest+=2;
  }
  break;
  
  case 16777216:
  if (vga_misordered)
  {
   if (vga_bytes==3)
   {
    /* 3 bytes misordered */
    for (;length;length--)
    {
     COMPUTE_INDICES
     *dest=r;
     dest[1]=g;
     dest[2]=b;
     dest+=3;
    }
   }
   else
   {
    /* 4 bytes misordered */
    for (;length;length--)
    {
     COMPUTE_INDICES
     dest[1]=b;
     dest[2]=g;
     dest[3]=r;
     dest+=4;
    }
   }
  }
  else
  {
   if (vga_bytes==3)
   {
    /* 3 bytes ordered */
    for (;length;length--)
    {
     COMPUTE_INDICES
     *dest=b;
     dest[1]=g;
     dest[2]=r;
     dest+=3;
    }
   }
   else
   {
    /* 4 bytes ordered */
    for (;length;length--)
    {
     COMPUTE_INDICES
     *dest=b;
     dest[1]=g;
     dest[2]=r;
     dest+=4;
    }
   }
  }
  break;
 }
}

/* memcpy. Copies pixel data into the screen. length is in pixels. */
void
transcpy(unsigned char *dest,const unsigned char *from, int length)
{
 memcpy(dest,from,length*vga_bytes);
}

/* Is optimized so that it switches the page only if it is not already
 * switched on the right value.
 */
void inline
switch_page(int page)
{
 if (active_page!=page)
 {
  active_page=page;
  vga_setpage(page);
 }
}

/* Length tells the number of pixels */
void
emit_stream(unsigned long *screen_offset, unsigned char **source, unsigned
 length)
{
 unsigned chunk_length,out_chunk_length;
 unsigned paga,page;
 unsigned char tmp[4];
 
 if (vga_linear)
 {
  transcpy(my_graph_mem+*screen_offset,*source,length);
  *screen_offset+=length*vga_bytes;
  *source+=length*vga_bytes;
 }
 else
 {
  while(length)
  {
   paga=*screen_offset&65535;
   page=*screen_offset>>16;
   switch_page(page);
   if (65536-paga<vga_bytes)
   {
    /* The ugly case on the boundary of the pages
     * We are going to generate exactly one pixel.
     */
    int i;
    transcpy(tmp,*source,1);
    for (i=0;i<65536-paga;i++,paga++)
     my_graph_mem[paga]=tmp[i];
    paga=0;
    page++;
    switch_page(page);
    for (;i<vga_bytes;i++,paga++)
     my_graph_mem[paga]=tmp[i];
    *source+=vga_bytes;
    length--;
    *screen_offset+=vga_bytes;
   }
   else
   {
    /* The nice case */
    out_chunk_length=min(vga_bytes*length,65536-paga);
    chunk_length=out_chunk_length/vga_bytes;
    transcpy(my_graph_mem+paga,*source,chunk_length);
    *source+=chunk_length*vga_bytes;
    *screen_offset+=chunk_length*vga_bytes;
    length-=chunk_length;
   }
  }
 }
}

/* Length tells the number of pixels
 * If flag is 1, emits foreground instead of background
 */
void
emit_background_stream(unsigned long *screen_offset, unsigned length)
{
 unsigned chunk_length,out_chunk_length;
 unsigned paga,page;
 unsigned char tmp[4];
 
 if (vga_linear)
 {
  background_transcpy(my_graph_mem+*screen_offset,length);
  *screen_offset+=length*vga_bytes;
 }
 else
 {
  while(length)
  {
   paga=*screen_offset&65535;
   page=*screen_offset>>16;
   switch_page(page);
   if (65536-paga<vga_bytes)
   {
    /* The ugly case on the boundary of the pages
     * We are going to generate exactly one pixel.
     */
    int i;
    background_transcpy(tmp,1);
    for (i=0;i<65536-paga;i++,paga++)
     my_graph_mem[paga]=tmp[i];
    paga=0;
    page++;
    switch_page(page);
    for (;i<vga_bytes;i++,paga++)
     my_graph_mem[paga]=tmp[i];
    length--;
    *screen_offset+=vga_bytes;
   }
   else
   {
    /* The nice case */
    out_chunk_length=min(vga_bytes*length,65536-paga);
    chunk_length=out_chunk_length/vga_bytes;
    background_transcpy(my_graph_mem+paga,chunk_length);
    *screen_offset+=chunk_length*vga_bytes;
    length-=chunk_length;
   }
  }
 }
}

/* Length tells the number of pixels. Updates screen_offset. Updates source.
 */
void
emit_text_stream(unsigned long *screen_offset, unsigned char **source, unsigned
 length)
{
 unsigned chunk_length,out_chunk_length;
 unsigned paga,page;
 unsigned char tmp[4];
 
 if (vga_linear)
 {
  text_transcpy(my_graph_mem+*screen_offset,*source,length);
  *screen_offset+=length*vga_bytes;
  *source+=length*vga_bytes;
 }
 else
 {
  while(length)
  {
   paga=*screen_offset&65535;
   page=*screen_offset>>16;
   switch_page(page);
   if (65536-paga<vga_bytes)
   {
    /* The ugly case on the boundary of the pages
     * We are going to generate exactly one pixel.
     */
    int i;
    text_transcpy(tmp,*source,1);
    for (i=0;i<65536-paga;i++,paga++)
     my_graph_mem[paga]=tmp[i];
    paga=0;
    page++;
    switch_page(page);
    for (;i<vga_bytes;i++,paga++)
     my_graph_mem[paga]=tmp[i];
    (*source)++;
    length--;
    *screen_offset+=vga_bytes;
   }
   else
   {
    /* The nice case */
    out_chunk_length=min(vga_bytes*length,65536-paga);
    chunk_length=out_chunk_length/vga_bytes;
    text_transcpy(my_graph_mem+paga,*source,chunk_length);
    *source+=chunk_length;
    *screen_offset+=chunk_length*vga_bytes;
    length-=chunk_length;
   }
  }
 }
}

/* Writes text into sl. sl can contain max. xsize characters. */
void
write_string_into_sl(int x, const unsigned char *txt,int length)
{
 if (x>=xsize) return;
 if (length>xsize-x) length=xsize-x;
 memcpy(sl+x,txt,length);
 if (sl_length<x+length)sl_length=x+length;
}

int
text_width(unsigned char *txt,int length)
{int ctr;
 unsigned char *t;
 int total;
 
 for (ctr=length,t=txt,total=0;ctr;ctr--,t++)
  total+=font_widths[*t];

 return total;
}

/* Sets up black background */
void
set_black(void)
{
 bgcolor_r=bgcolor_g=bgcolor_b=0;
}

/* Returns the offset of which the start was increased. */
int
adjust_start_length(int *start, int *length, int begin, int end)
{
 int offset=0;
 
 if (*length<0) goto empty;
 if (*start+*length<=begin) goto empty;
 if (*start>=end) goto empty;
 if (*start<begin)
 {
  offset=begin-*start;
  *start=begin;
  *length-=offset;
 }
 if (*start+*length>end)
 {
  *length=end-*start;
 }
 return offset;
 
 empty:
 *length=0;
 return offset;
}

/* Writes a line (from "lines") 
   one pixel high, source is internal pixel representation.
   length is length of the line in pixels.
   line is y-coordinate of the line on the screen.
   Fills in the remainder of the line with black.
*/
void
write_line_to_screen(unsigned char *src,int length,int y)
{int x=-x_off;
 unsigned long lina; /* Linear screen address */
 unsigned char *srcptr; /* Scans src */
 int remains; /* Number of pixels of current scanline to write */
 
 if (y<0||y>=ysize) return;

 src+=vga_bytes*adjust_start_length(&x,&length,0,xsize);
 lina=vga_linewidth*y;
 srcptr=src+x*vga_bytes;
 remains=length;
 if (remains>xsize) remains=xsize;
 emit_stream(&lina,&srcptr,remains);
 if (global_no_tails) return;
 remains=xsize-remains;
 set_black();
 bgtrans_flag=0;
 emit_background_stream(&lina,remains);
}

/* Draws a font_height high strip with letters 
 * y is in pixels, not in chars. */
void
write_text_line_to_screen(unsigned char *source,unsigned length,int *x0, int y)
{
 unsigned char *font_slice=font_img;
 unsigned long screen_offset;
 unsigned char *txt_ptr;
 unsigned txt_ctr;
 int scanlines=font_height,scanline;
 int start_scanline=y;
 int remains,x;
  
 font_slice+=font_width*adjust_start_length(&start_scanline,&scanlines,0,ysize);
 
 while(length&&font_widths[*source]+*x0<=0)
 {
  length--;
  *x0+=font_widths[*source++];
 }
 if (!length) return;
 x=*x0;
 for (scanline=start_scanline;scanline<start_scanline+scanlines;scanline++)
 {
  x=*x0;
  txt_ptr=source;
  txt_ctr=length;
  if (x<0)
  {
   unsigned char *dummy=font_slice+font_stops[*txt_ptr]-x;
    
   remains=font_widths[*txt_ptr]+x;
   screen_offset=scanline*vga_linewidth;
   if (remains+x>xsize) remains=xsize-x;
   emit_text_stream(&screen_offset,&dummy,remains);
   x=0;
   txt_ptr++;
   txt_ctr--;
  }
  else
  {
   screen_offset=scanline*vga_linewidth+x*vga_bytes;
  }
  while(x<xsize&&txt_ctr)
  {
   remains=font_widths[*txt_ptr];
   if (x+remains>xsize) remains=xsize-x;
   {
    unsigned char *dummy=font_slice+font_stops[*txt_ptr];
    emit_text_stream(&screen_offset,&dummy,remains);
   }
   txt_ctr--;
   txt_ptr++;
   x+=remains;
  }
  font_slice+=font_width;
 } 
 *x0=x;
}

/* Like write_text_line_to_screen, but draw only background color
 * y is not in text lines, but in scanlines. The output ends at the right edge
 * of the screen.
 * Uses global variables filler_start and filler_end to determine which
 * scanlines should be filled with background and which with foreground. 
 */
void
write_background_foreground_line_to_screen(int *x, int y, int length)
{
 unsigned long screen_offset;
 int scanline;
 int scanlines=font_height;
 int start_scanline=y;

 if (*x>=xsize) return;
 if (*x<0)
 {
  length+=*x;
  *x=0;
 }
 if (*x+length>xsize) length=xsize-*x;
 adjust_start_length(&start_scanline,&scanlines,0,ysize);
 for (scanline=0;scanline<scanlines;scanline++)
 {
  screen_offset=(scanline+start_scanline)*vga_linewidth+*x*vga_bytes;
  bgtrans_flag=(scanline>=filler_start_scanline&&scanline<filler_end_scanline);
  emit_background_stream(&screen_offset,length);
 }
 *x+=length;
}

void
write_background_line_to_screen(int x, int y)
{
 int length=xsize-x;

 filler_start_scanline=filler_end_scanline=0;
 write_background_foreground_line_to_screen(&x,y,length);
}

int
m_to_int(unsigned char *src)
{
 return *src+(src[1]<<8)+(src[2]<<16)+(src[3]<<24);
}

/* Returns a buffer with the broken down chars. All box-like things are replaced
 * with 256 which means the automatus should be reset.
 * Stores the result in broken_down_line and broken_down_line_length
 */
void
break_down_line(bd src)
{
 struct char_desc *broken_down_line_ptr;
 unsigned char *code=src.ptr;
 int code_length=src.number;
 int l;
 
 /* First, we have to count the number of chars s_line_length to store the line */
 free(broken_down_line); 
 while(code_length>0)
 {
  switch(*code)
  {
   case 0:
   l=code[1];
   broken_down_line_length+=l;
   code+=l+2;
   code_length-=l+2;
   break;

   case 1:
   l=m_to_int(code+1);
   broken_down_line_length+=l;
   code+=l+5;
   code_length-=l;
   break;
   
   case 4:
   code+=2;
   code_length-=2;
   broken_down_line_length++;
   break;

   case 5:
   code+=5;
   code_length-=5;
   broken_down_line_length++;
   break;
  }
 }
 broken_down_line=xalloc(broken_down_line_length*sizeof(*broken_down_line));
 broken_down_line_ptr=broken_down_line;
 code=src.ptr;
 code_length=src.number;
 while(code_length>0)
 {
  switch(*code)
  {
   case 0:
   l=code[1];
   {
    int a;
    for (a=0;a<l;a++)
    {
     broken_down_line[a].code=code[2+a];
     broken_down_line[a].type=5;
     broken_down_line[a].br=bgcolor_r;
     broken_down_line[a].bg=bgcolor_g;
     broken_down_line[a].bb=bgcolor_b;
     broken_down_line[a].fr=fgcolor_r;
     broken_down_line[a].fg=fgcolor_g;
     broken_down_line[a].fb=fgcolor_b;
    }
   }
   broken_down_line_ptr+=l;
   code+=2+l;
   code_length-=l+2;
   break;

   case 1:
   l=m_to_int(code+1);
   {
    int a;
    for (a=0;a<l;a++)
    {
     broken_down_line[a].code=code[5+a];
     broken_down_line[a].type=5;
     broken_down_line[a].br=bgcolor_r;
     broken_down_line[a].bg=bgcolor_g;
     broken_down_line[a].bb=bgcolor_b;
     broken_down_line[a].fr=fgcolor_r;
     broken_down_line[a].fg=fgcolor_g;
     broken_down_line[a].fb=fgcolor_b;
    }
   }
   broken_down_line_ptr+=l;
   code+=5+l;
   code_length-=l+5;
   break;

   case 2: /* bgcolor. 2, red, green, blue */
   bgcolor_r=code[1];
   bgcolor_g=code[2];
   bgcolor_b=code[3];
   code+=4;
   code_length-=4;
   break;
   
   case 3: /* fgcolor. 3, red, green, blue */
   fgcolor_r=code[1];
   fgcolor_g=code[2];
   fgcolor_b=code[3];
   code+=4;
   code_length-=4;
   break;

   case 4:
   broken_down_line_ptr->type=code[1];
   broken_down_line_ptr->code=code[2];
   broken_down_line->br=bgcolor_r;
   broken_down_line->bg=bgcolor_g;
   broken_down_line->bb=bgcolor_b;
   broken_down_line->fr=fgcolor_r;
   broken_down_line->fg=fgcolor_g;
   broken_down_line->fb=fgcolor_b;
   code+=2;
   code_length-=2;
   break;

   case 5:
   broken_down_line_ptr->type=code[1];
   broken_down_line_ptr->code=m_to_int(code+2);
   broken_down_line->br=bgcolor_r;
   broken_down_line->bg=bgcolor_g;
   broken_down_line->bb=bgcolor_b;
   broken_down_line->fr=fgcolor_r;
   broken_down_line->fg=fgcolor_g;
   broken_down_line->fb=fgcolor_b;
   code+=5;
   code_length-=5;
   break;
  }
 }
}

/* On text line number y and x offset -x draws the output from the code.
 * If the code is corrupted, segfault results!
 * If image_mode, then no interpreting is performed
 * Do not call this in image mode.
 */
void
interpret_line(unsigned char *code, int code_length, int y)
{
 int length,x;
 
 if (image_mode)
 {
  write_line_to_screen(code,code_length,y);
  return;
 }
 x=-x_off;
 while (code_length>0&&x<xsize)
 {
  switch(*code)
  {
   case 0: /* Short text. 0, length, text */
   length=code[1];
   write_text_line_to_screen(code+2,length,&x,y*font_height);
   code_length-=length+2;
   code+=length+2;
   break;
   
   case 1: /* Long text. 1, length, length, length, length, text */
   length=m_to_int(code+1);
   write_text_line_to_screen(code+5,length,&x,y*font_height);
   code_length-=length+5;
   code+=length+5;
   break;
   
   case 2: /* bgcolor. 2, red, green, blue */
   bgcolor_r=code[1];
   bgcolor_g=code[2];
   bgcolor_b=code[3];
   code+=4;
   code_length-=4;
   break;
   
   case 3: /* fgcolor. 3, red, green, blue */
   fgcolor_r=code[1];
   fgcolor_g=code[2];
   fgcolor_b=code[3];
   code+=4;
   code_length-=4;
   break;

   case 4: /* Short filler. 4, type, length.
              Type: 0=background
                    1=top half including horizontal line filled
                    2=bottom half including horiz line filled
                    3=completely filled
                    4=horizontal line filled */
   length=code[2];
   general_filler:
   switch(code[1])
   {
    /* Foregrounded are the scalines s for which (s>=filler_start_scanline
     * && s<filler_end_scanline) holds.
     */
    case 0: 
    filler_start_scanline=filler_end_scanline=0;
    break;
    
    case 1:
    filler_start_scanline=0;
    filler_end_scanline=(font_height+box_horiz_thickness)/2;
    break;

    case 2:
    filler_start_scanline=(font_height-box_horiz_thickness)/2;
    filler_end_scanline=font_height;
    break;

    case 3:
    filler_start_scanline=0;
    filler_end_scanline=font_height;
    break;
    
    case 4:
    filler_start_scanline=(font_height-box_horiz_thickness)/2;
    filler_end_scanline=(font_height+box_horiz_thickness)/2;
    break;
   }
   write_background_foreground_line_to_screen(&x,y*font_height,length);
   if (*code==5)
   {
    code+=5;
    code_length-=5;
   }
   else
   {
    code+=2;
    code_length-=2;
   }
   break;

   case 5: /* Long filller. 5, type, length, length, length, length */
   length=m_to_int(code+2);
   goto general_filler;
   break;
  }
 }
 write_background_line_to_screen(x,y*font_height);
}

void
fprint_block(FILE *f, unsigned char *source, int length)
{
 int a;
 for (a=0;a<length;a++)
  fputc(source[a],f);
 fputc('\n',f);
 fprintf(stderr,"End of fprint_block\n");
}

void
set_sl_colors(void)
{
 bgcolor_r=slbgcolor_r;
 bgcolor_g=slbgcolor_g;
 bgcolor_b=slbgcolor_b;
 fgcolor_r=slfgcolor_r;
 fgcolor_g=slfgcolor_g;
 fgcolor_b=slfgcolor_b;
}

void
blit_sl(void)
{
 int dummy;
 dummy=0;
 
 set_sl_colors();
 write_text_line_to_screen(sl,sl_length,&dummy,image_mode?top_lines:
  top_lines*font_height);
 write_background_line_to_screen(dummy,image_mode?top_lines:
  top_lines*font_height);
}

/* lines_pos tells how many lines were filled in, i.e., space was allocd
   for their characters. */
void
free_all_lines()
{int a;
 struct block_desc *l=lines;
 
 if (!lines) 
 {
  lines_pos=0;
  lines_allocd=0;
  return;
 }
 for (a=lines_pos;a;a--)
 {free(l->ptr);
  l++;
 }
 free(lines);
 lines_pos=0;
 lines_allocd=0;
 lines=NULL;
}

/* p is a pointer to 768 doubles defining the 256 entries of the palette. */
void 
set_0_1_palette(double *p)
{
 Palette o;
 int n;

 for(n=0;n<768;n++)
 {
  if (p[n]<0) p[n]=0;
  if (p[n]>1) p[n]=1;
 }

 for (n=0;n<256;n++)
 {unsigned a;
  a=floor(.5+63*p[n*3]);o.color[n].red=a;
  a=floor(.5+63*p[n*3+1]);o.color[n].green=a;
  a=floor(.5+63*p[n*3+2]);o.color[n].blue=a;
 }
 gl_setpalette(&o);
}

/* Loads the 3-3-2 256-color cube palette without gamma correction
   Memory layout: bits 76543210
                       RRRGGGBB
*/
void                       
setup_256_color_palette(void)
{
 double p[768],*pp=p;
 double r,g,b;
 int i;

 for (i=0;i<256;i++)
 {
  r=(double)(i>>5)/7;
  g=(double)((i>>2)&7)/7;
  b=(double)(i&3)/3;
  *pp=r;
  pp[1]=g;
  pp[2]=b;
  pp+=3;
 }
 set_0_1_palette(p);
}

/* Makes n entries in the lut and for each tells the linear intensity, 0
   is black and 65536 is full intensity. */
int
compute_dither_lut(int **lut, int n, double gamma, double mul, double shift)
{
 int i;
 double v;
 
 free(*lut);
 *lut=xalloc(n*sizeof(**lut));
 for (i=0;i<n;i++)
 {
  v=(double)i/(n-1);
  v=(v*mul)+shift;
  if (v<0) v=0;
  v=pow(v,gamma);
  v=v*PRESCALE+.5;
  if (v>(1<<30)) v=(1<<30);
  (*lut)[i]=floor(v);
 }
 return 0;
}

int inline
lookup(int *lut, int bit,int val)
{
 int i=0,i1,v,v1;
 int mask=1<<bit,n_table=mask;
 
 for (;bit;bit--)
 {
  mask>>=1;
  i|=(lut[i|mask]<=val)?mask:0;
 }
 i1=i+1;
 v=lut[i];
 if (i1>=n_table) return i;
 v1=lut[i1];
 if (v1-val>val-v) return i; else return i1;
}

/* Emits VGA repsesentation of rounding of the pixel and the remainder which
 * wasn't painted leaves in the r,g,b. Remainder can be symmetrically negative
 * or positive. r,g,b are from -PRESCALE (-black) to PRESCALE (black).
 */
void
round_pixel(unsigned char *dest, int *r, int *g, int *b)
{
 int o;
 int ri=lookup(dither_lut_r,dither_e_r,*r);
 int gi=lookup(dither_lut_g,dither_e_g,*g);
 int bi=lookup(dither_lut_b,dither_e_b,*b);
 
 *r-=dither_lut_r[ri];
 *g-=dither_lut_g[gi];
 *b-=dither_lut_b[bi];
 switch(vga_colors)
 {
  case 256: o=(ri<<5)|(gi<<2)|bi; break;
  case 32768: o=(ri<<10)|(gi<<5)|bi; break;
  case 65536: o=(ri<<11)|(gi<<5)|bi; break;
  default: /* 16777216 */
  if (vga_bytes==3)
   if (vga_misordered)
    o=(bi<<16)|(gi<<8)|ri;
   else
    o=(ri<<16)|(gi<<8)|bi;
  else
   if (vga_misordered)
    o=(ri<<24)|(gi<<16)|(bi<<8);
   else
    o=(ri<<16)|(gi<<8)|bi;
  break;
 }
 switch (vga_bytes)
 {
  case 1: *dest=o; break;
  case 2: *dest=o; dest[1]=o>>8; break;
  case 3: *dest=o; dest[1]=o>>8; dest[2]=o>>16; break;
  case 4: *dest=o; dest[1]=o>>8; dest[2]=o>>16; dest[3]=o>>24; break;
 }
}

/* Luts have each 256 entries.
 * Gamma tells the gamma value of the monitor on that channel
 * The input (linear) color is multiplied by mul, shift is added and
 * the result is powered to gamma^-1. Then the best approximation of ths
 * voltage is found. n tells the last color index for the given channel.
 * For example, for 3-3-2 palette and R channel, the n is 7.
 */
int
compute_gamma_lut(unsigned char *lut,unsigned n,double gamma,double mul,double
shift)
{
 int vv,i;
 double v,gg;
 
 gg=1/gamma;
 for (i=0;i<256;i++)
 {
  v=(double)i/255;
  v=pow(v,gg);
  v-=shift;
  v/=mul;
  vv=floor(v*(n-1)+.5);
  if (vv<0) vv=0;
  if (vv>n-1) vv=n-1;
  lut[i]=vv;
 }
 return 0;
}

int
setup_color_curves(void)
{
 if (monitor_gamma_r<=0||monitor_gamma_g<=0||monitor_gamma_b<=0)
 {
  fprintf(stderr,"Invalid monitor gamma in one the channels.\n");
  return 1;
 }
 if (monitor_mul_r<=0||monitor_mul_g<=0||monitor_mul_b<=0)
 {
  fprintf(stderr,"Invalid monitor mul in one of the channels.\n");
  return 1;
 }
 if (vga_colors==256)
 {
  setup_256_color_palette();
 }
 {
  switch(vga_colors)
  {
   case 256: dither_e_r=dither_e_g=3;dither_e_b=2; break;
   case 32768:dither_e_r=dither_e_g=dither_e_b=5;break;
   case 65536:dither_e_r=dither_e_b=5;dither_e_g=6;break;
   default:dither_e_r=dither_e_g=dither_e_b=8;break;
  }
  dither_n_r=1<<dither_e_r;
  dither_n_g=1<<dither_e_g;
  dither_n_b=1<<dither_e_b;
  if (compute_gamma_lut(gamma_lut_r,dither_n_r,
   monitor_gamma_r/required_gamma,monitor_mul_r,monitor_shift_r)) return 1;
  if (compute_dither_lut(&dither_lut_r,dither_n_r,
   monitor_gamma_r/required_gamma,monitor_mul_r,monitor_shift_r)) return 1;
  if (compute_gamma_lut(gamma_lut_g,dither_n_g,
   monitor_gamma_g/required_gamma,monitor_mul_g,monitor_shift_g)) return 1;
  if (compute_dither_lut(&dither_lut_g,dither_n_g,
   monitor_gamma_g/required_gamma,monitor_mul_g,monitor_shift_g)) return 1;
  if (compute_gamma_lut(gamma_lut_b,dither_n_b,
   monitor_gamma_b/required_gamma,monitor_mul_b,monitor_shift_b)) return 1;
  if (compute_dither_lut(&dither_lut_b,dither_n_b,
   monitor_gamma_b/required_gamma,monitor_mul_b,monitor_shift_b)) return 1;
 }
 return 0;
}

/* Makes the "sl" to contain just the given message and redraws its contents
 * onto the screen.
 */
void
message(const unsigned char *t)
{
 sl_length=0;
 write_string_into_sl(0,t,strlen(t));
 blit_sl();
}

/* Prints line number statistics. */
void
print_line_count(void)
{
 unsigned l,t,p;
 unsigned char txt[32];
 
 #ifdef DG_SCROLLER
  fprintf(stderr,"I got into print_line_count.\n");
 #endif
 l=(unsigned) sc_b_line;
 l%=999999;
 l++;
 t=(unsigned) lines_pos;
 t%=1000000;
 if (lines_pos<=top_lines)
 {
  p=100;
 }
 else
 {
  p=sc_b_line*100/(lines_pos-top_lines);
 }
 p%=1000;
 sprintf(txt,"Line %u/%u %u%% ",l,t,p);
 message(txt);
}

/* Clears the lines of screen (zeroes it out)
   input is in scanlines. */
void
clear_screen (int startline,int lines)
{
 unsigned long screen_offset;
 
 set_black();
 adjust_start_length(&startline,&lines,0,ysize);
 lines+=startline;
 for (;startline<lines;startline++)
 {
  screen_offset=startline*vga_linewidth;
  bgtrans_flag=0;
  emit_background_stream(&screen_offset,xsize);
 }
 getchar();
}
  
/* Clears the space between sl and bottom of the screen. */
void
clear_under_sl(void)
{
 int a;
 a=top_lines+font_height;
 clear_screen(a,ysize-a);
}

/* Redraws the top part of the screen. */
void
redraw_screen(void)
{
 int height=top_lines;
 int y;

 if (image_mode)
 {
  if (sc_b_line+height>lines_pos)
   height=lines_pos-sc_b_line;
  for (y=0;y<height;y++)
   write_line_to_screen(lines[sc_b_line+y].ptr,lines[sc_b_line+y].number,y);  
  clear_screen(y,top_lines-y);
 }
 else
 {
  if (sc_b_line+height>lines_pos)
   height=lines_pos-sc_b_line;
  for (y=0;y<height;y++)
   interpret_line(lines[sc_b_line+y].ptr,lines[sc_b_line+y].number,y);
  set_black();
  for (;y<top_lines;y++)
   write_background_line_to_screen(0,y*font_height);
 }
}

/* It adds the line into the line list.
   When in appropriate range, line is also printed onto the screen.
   When in image mode, adds pixels instead of chars. :-)
   ln->number means number of chars in text mode and number of PIXELS,
   not bytes in image mode. The number of bytes in image mode is
   ln->number*vga_bytes.
   return value 0 ok
                1 fatal error
*/
int
add_lineno(struct block_desc *ln,int lineno)
{
 int bytes_per_unit=image_mode?vga_bytes:1;

 #ifdef DG_SCROLLER
  fprintf(stderr,"add_lineno, len=%i, lineno=%i, lines_pos=%i, lines_allocd=%i.\n",
   ln->number,lineno,lines_pos,lines_allocd);
 #endif
 if (lineno<0) return 0; /* Nonsense */
 /* First, we have to test if there are enough lines */
 if (lineno>=lines_allocd)
 {
 /* We have to alloc new lines */
  lines_allocd=64+lineno; /* A magic number :-) */
  #ifdef DG_SCROLLER
   fprintf(stderr,"lines_allocd ->%d.\n",lines_allocd);
  #endif
  lines=(struct block_desc *)realloc(lines,lines_allocd*sizeof(struct
   block_desc));
  if (!lines)
  {
   fprintf(stderr,"SC: Unable to realloc line buffers.\n");
   return 1;
  }
 }
 /* Then we have to zero out the intermediate lines */
 {
  int a;
  for (a=lines_pos;a<lineno;a++)
  {
   lines[a].ptr=NULL;
   lines[a].number=0;
  }
 }
 #ifdef DG_SCROLLER
  fprintf(stderr,"Intermediate zeroed.\n");
 #endif
 /* If the line was in use, we have to free it up: */
 if (lineno<lines_pos)
 {
  free(lines[lineno].ptr);
 }
 #ifdef DG_SCROLLER
  fprintf(stderr,"Going to malloc.\n");
 #endif
 /* Next we have to alloc space for the new line to be copied. */ 
 if (!(lines[lineno].ptr=malloc(ln->number*bytes_per_unit))&&ln->number)
 {
  fprintf(stderr,"SC: Unable to malloc space for incoming micro-line.\n");
  return 1;
 }
 #ifdef DG_SCROLLER
  fprintf(stderr,"Malloced %d bytes for incoming lineno.\n",
   ln->number*bytes_per_unit);
 #endif
 memcpy(lines[lineno].ptr,ln->ptr,
  ln->number*bytes_per_unit); /* Copy line chars, there */
 /* are no 0 terminators */
 #ifdef DG_SCROLLER
  fprintf(stderr,"After memcpy.\n");
 #endif
 lines[lineno].number=ln->number;
 if (lineno>=sc_b_line&&lineno<sc_b_line+top_lines)
 {
  /* micro will be written onscreen */
  if (image_mode)
  {
   write_line_to_screen(lines[lineno].ptr,lines[lineno].number,
    lineno-sc_b_line);
  }
  else
  {
   /* text will be written onscreen and into vs */
   interpret_line(lines[lines_pos].ptr,lines[lines_pos].number,
    lines_pos-sc_b_line);
  }
 }
 if (lineno+1>lines_pos)
 {
  lines_pos=lineno+1;
 }
 if (image_mode&&ln->number>max_width)
 {
  max_width=ln->number;
 }
 return 0;
}

/* Adds text line by means of breaking it into the pixels and adding
   the resulting micro lines.
   return value 0 ok
                1 fatal error
*/
int
add_pixelize_line(struct lineno_desc *ln)
{
 int y;
 unsigned char *tmpptr; /* Pointer into tmp */
 unsigned char *fline=font_img; /* The beginning of the font line */
 unsigned char *tptr; /* Pointer into ascii codes of the characters */
 int ctr; /* Counter of the chars of the text */
 struct block_desc block;
 
 block.number=text_width(ln->ptr,ln->number);
 block.ptr=(unsigned char *)malloc(block.number*vga_bytes);
 if (block.number&&!block.ptr)
 {
  fprintf(stderr,"Unable to malloc %i bytes for tmp buffer in"
   " add_pixelize_line.\n",block.number*vga_bytes);
  return 1;
 }
 for (y=font_height;y;y--)
 {
  tptr=ln->ptr;
  tmpptr=block.ptr;
  for (ctr=ln->number;ctr;ctr--)
  {
   text_transcpy(tmpptr,fline+font_stops[*tptr],font_widths[*tptr]);
   tmpptr+=vga_bytes*font_widths[*tptr];
   tptr++;
  }
  add_lineno(&block,ln->lineno+font_height-y);
  fline+=font_width;
 }
 free(block.ptr);
 return 0;
}

void
up(unsigned m)
{
 if (m>sc_b_line) m=sc_b_line;
 if (!m) return;
 sc_b_line-=m;
 #ifdef DG_SCROLLER
  fprintf(stderr,"Calling redraw_screen.\n");
 #endif
 redraw_screen();
 print_line_count();
}

void
down(unsigned m)
{
 if (!lines_pos) return;
 if (sc_b_line+m>=lines_pos)
 {
  m=lines_pos-1-sc_b_line;
 }
 if (!m) return;
 sc_b_line+=m;
 #ifdef DG_SCROLLER
  fprintf(stderr,"Calling redraw_screen.\n");
 #endif
 redraw_screen();
 print_line_count();
}

/* Comprises both up and down: 0 nothing
                              <0 up
                              >0 down
*/
void
move(int offset)
{
 if (!offset) return;
 if (offset>0)
 {
  down(offset);
 }
 else
 {
  up(-offset);
 }
}

/* Prints a text line on the printer. */
void
print_text_line(unsigned char *ptr, int chars)
{
 int scanline,x;
 unsigned char *pixelptr,*charptr,*pixels,*fontline;
 int width,charctr;
 
 charptr=ptr;
 width=0;
 for (charctr=chars;charctr;charctr--)
 {
  width+=font_widths[*charptr++];
 }
 pixels=malloc(width);
 if (!pixels) return;
 fontline=font_img;
 for (scanline=font_height;scanline;scanline--)
 {
  charptr=ptr;
  pixelptr=pixels;
  for (charctr=chars;charctr;charctr--)
  {
   memcpy(pixelptr,fontline+font_stops[*charptr],font_widths[*charptr]);
   pixelptr+=font_widths[*charptr];
   charptr++;
  }
  fontline+=font_width;
  for (x=0;x<width;x++)
  {
   pixels[x]&=63;
   pixels[x]=(pixels[x]>=10);
  }
  add_printer_line(pixels,width);
 }
 free(pixels);
}

/* Prints the contents of the whole page. */
void
print_contents(void)
{
 int mline,line,width,x;
 unsigned char *ptr,*tmp;
 
 if (!image_mode)
 {
  /* Text mode */
  for (line=0;line<lines_pos;line++)
  {
   print_text_line(lines[line].ptr,lines[line].number);
  }
 }
 else
 {
  /* Graphics mode */
  for (mline=0;mline<lines_pos;mline++)
  {
   width=lines[mline].number;
   ptr=lines[mline].ptr;
   tmp=malloc(width);
   if (!tmp) return;
   for (x=0;x<width;x++)
    tmp[x]=1;
   add_printer_line(tmp,width);
   free(tmp);
  }
 }
}

void
print_mode_info(void)
{
 fprintf(stderr,"Graphics mode info\n------------------\n");
 fprintf(stderr,"xsize=%d, ysize=%d, my_graph_mem=%p\n",xsize,ysize,my_graph_mem);
 fprintf(stderr,"vga_linewidth=%d, vga_linear=%d, vga_misordered=%d\n",
  vga_linewidth, vga_linear, vga_misordered);
 fprintf(stderr,"vga_bytes=%d vga_colors=%d\n\n", vga_bytes, vga_colors);
}

void
push_and_paint(int start_line, int start_char, int end_line, int end_char)
{
}

/* Scroller's main routine. */
int
scroller(int command,void *params)
{
 struct block_desc bd;
 struct lineno_desc ln;
 /* lines_allocd tells how many lines were param to realloc(lines) */
 /* If lines==NULL&&lines_allocd==0, it means that there are no lines. */
 /* top_lines tells how much lines is visible. Changes with every scren
      or font change. When 25 lines fit onscreen and one is used as a
      status line, top_lines=24.*/
 /* bd is the param passed by the caller. */
  
 switch(command)
 {
  case SC_NEW_SCREEN:
  image_mode=*(int *)params;
  #ifdef DG_SCROLLER
   fprintf(stderr,"I got into SC_NEW_SCREEN. image_mode=%i\n",image_mode);
  #endif
  if (image_mode)
  {
   x_step=xsize/8;
   max_width=0; /* The lines are going to be freed */
   top_lines=ysize-font_height;
  }
  else
  {
   top_lines=ysize/font_height-1; /* One less for status line */
  }
  #ifdef DG_SCROLLER
   fprintf(stderr,"I got into SC_NEW_SCREEN-1.\n");
  #endif
  if (top_lines<1) top_lines=1;
  if (image_mode)
  {
   /* Image mode */
   page_skip=top_lines/4;
  }
  else
  {
   /* Text mode */
   page_skip=top_lines; /* No lines will be in intersection of
                            adjoining pages. */
  }
  #ifdef DG_SCROLLER
   fprintf(stderr,"I got into SC_NEW_SCREEN-2.\n");
  #endif
  if (page_skip<1) page_skip=1;
  free_all_lines();
  sl_length=0;
  free(sl);
  sl=calloc(xsize,1);
  #ifdef DG_SCROLLER
   fprintf(stderr,"I got into SC_NEW_SCREEN-3.\n");
  #endif
  if (!sl)
  {
   fprintf(stderr,"Unable to malloc %i bytes for status line.\n",xsize);
   return 1;
  }
  if (image_mode)
  {
   /* Image mode */
   clear_screen(0,top_lines);/* Clear the top part of the screen */
   x_off=0;
  }
  else
  {
   /* Text mode */
   clear_screen(0,top_lines*font_height);
   /* Clear the top part of the screen */
  }
  #ifdef DG_SCROLLER
   fprintf(stderr,"I got into SC_NEW_SCREEN-4.\n");
  #endif
  blit_sl();
  #ifdef DG_SCROLLER
   fprintf(stderr,"I got into SC_NEW_SCREEN-5.\n");
  #endif
  clear_under_sl();
  #ifdef DG_SCROLLER
   fprintf(stderr,"I got into SC_NEW_SCREEN-6.\n");
  #endif
  #ifdef DG_SCROLLER
   fprintf(stderr,"Ending SC_NEW_SCREEN.\n");
  #endif
  break;
  
  case SC_NEW_FONT:
  #ifdef DG_SCROLLER
   fprintf(stderr,"I got into SC_NEW_FONT.\n");
  #endif
  if (image_mode)
  {
   top_lines=ysize-font_height;
  }
  else
  {
   top_lines=ysize/font_height-1; /* One less for status line */
  }
  if (top_lines<1) top_lines=1;
  if (image_mode)
  {
   /* Image mode */
   page_skip=top_lines/4;
  }
  else
  {
   /* Text mode */
   page_skip=top_lines-0; /* No lines will be in the intersection of adjoinig */
                         /* lines */
  }
  if (page_skip<1) page_skip=1;
  if (!image_mode)
  {
   free_all_lines();
  }
  blit_sl();
  clear_under_sl();
  #ifdef DG_SCROLLER
   fprintf(stderr,"I completed SC_NEW_FONT.\n");
  #endif
  break;
  
  case SC_HOME:
  #ifdef DG_SCROLLER
   fprintf(stderr,"SC_HOME\n");
  #endif
  up(sc_b_line);
  break;

  case SC_END:
  #ifdef DG_SCROLLER
   fprintf(stderr,"SC_END\n");
  #endif
  move(lines_pos-top_lines-sc_b_line);
  break;
  
  case SC_LINE_UP:
  {
   unsigned m=*(unsigned *)params;
   
   #ifdef DG_SCROLLER
    fprintf(stderr,"SC_LINE_UP\n");
   #endif
   up(m);
  }
  break;
  
  case SC_LINE_DOWN:
  {
   unsigned m=*(unsigned *)params;
   
   #ifdef DG_SCROLLER
    fprintf(stderr,"SC_LINE_DOWN\n");
   #endif
   down(m);
  }
  break;
  
  case SC_PAGE_UP:
  {
   unsigned m;
   
   #ifdef DG_SCROLLER
    fprintf(stderr,"SC_PAGE_UP\n");
   #endif
   m=*(unsigned *)params;
   up(m*page_skip);
  }
  break;
  
  case SC_PAGE_DOWN:
  {
   unsigned m;

   #ifdef DG_SCROLLER
    fprintf(stderr,"SC_PAGE_DOWN\n");
   #endif
   m=*(unsigned *)params;
   down(m*page_skip);
  }
  break;
  
  case SC_ADD_MICRO_LINENO:
  /* The length of the parameter tells the number of pixels, not the number of
   * bytes! This call will be ignored in text mode.
   * Draws only the pixels, doesn't draw the black tail!
   */
  #ifdef DG_SCROLLER
   fprintf(stderr,"I got into SC_ADD_MICRO_LINENO.\n");
  #endif
  if (!image_mode) return 0; /* Nonsense. This is not allowed in text mode. */
  ln=*((struct lineno_desc *)params);
  bd.number=ln.number;
  bd.ptr=ln.ptr;
  global_no_tails=1;
  if (add_lineno(&bd,ln.lineno))
  {
   global_no_tails=0;
   fprintf(stderr,"Unable to add micro line.\n");
   return 1;
  }
  global_no_tails=0;
  break;
  
  case SC_ADD_MICRO_LINE:
  #ifdef DG_SCROLLER 
   fprintf(stderr,"I got into SC_ADD_MICRO_LINE.\n");
  #endif
  if (!image_mode) return 0; /* Nonsense */
  bd=*((struct block_desc *)params);
  if (add_lineno(&bd,lines_pos))
  {
   fprintf(stderr,"Unable to add micro line.\n");
   return 1;
  }
  #ifdef DG_SCROLLER
   fprintf(stderr,"I got out of SC_ADD_MICRO_LINE.\n");
  #endif
  break;
  
  case SC_ADD_LINENO:
  #ifdef DG_SCROLLER  
   fprintf(stderr,"I got into SC_ADD_LINENO, image_mode=%d\n",image_mode);
  #endif
  if (!image_mode) break;
  ln=*(struct lineno_desc *)params;
  set_sl_colors();
  #ifdef DG_SCROLLER
   fprintf(stderr,"I got before add_pixelize_line in ADD_LINENO.\n");
  #endif
  if (add_pixelize_line(&ln))
  {
   fprintf(stderr,"Can't SC_ADD_LINENO.\n");
   return 1;
  }
  break;
  
  case SC_ADD_LINE:
  /* When in image mode, it takes the line and parses it into pixels. */
  /* In textmode, it only remembers the codes of the characters.
   * In text mode, supply interpreter code. In image mode, supply only codes of
   * the chars.
   */
  #ifdef DG_SCROLLER  
   fprintf(stderr,"I got into SC_ADD_LINE, image_mode=%d\n",image_mode);
  #endif
  bd=*((struct block_desc *)params);
  if (image_mode)
  {
   /* Image mode */
   #ifdef DG_SCROLLER
    fprintf(stderr,"I got before add_pixelize_line.\n");
   #endif
   set_sl_colors();
   {
    struct lineno_desc ld;
    ld.ptr=bd.ptr;
    ld.number=bd.number;
    ld.lineno=lines_pos;
    if (add_pixelize_line(&ld))
    {
     fprintf(stderr,"Unable to add pixelized text line.\n");
     return 1;
    }
   }
  }
  else
  {
   /* Text mode */
   if (add_lineno(&bd,lines_pos))
   {
    fprintf(stderr,"Unable to add a line.\n");
    return 1;
   }
  }
  break;
 
  case SC_DISPLAY_LINE:                                    
  #ifdef DG_SCROLLER                                    
   fprintf(stderr,"scroller: I got into SC_DISPLAY_LINE.\n");
  #endif                                                   
  bd=*((struct block_desc *)params);                       
  sl_length=0;
  write_string_into_sl(0,bd.ptr,bd.number);                
  blit_sl();                                               
  break;                                                   
 
  case SC_LEFT:
  {
   unsigned m;
   m=*(unsigned *)params;
   if (x_off&&image_mode)
   {
    x_off-=x_step*m;
    if (x_off<0) 
    {
     x_off=0;
    }
    redraw_screen();
   }
  }
  break;

  case SC_RIGHT:
  {
   unsigned m;

   m=*(unsigned *)params;
   if (x_off+1>=max_width||!image_mode)
   {
    break; /* No action or text mode */
   }
   x_off+=m*x_step;
   if (x_off>=max_width)
   {
    x_off=max_width-1;
   }
   redraw_screen();
  }
  break;

  case SC_PRINT:
  if(open_printer())
  {
   fprintf(stderr,"scroller.c: unable to open printer\n");
   return 1;
  }
  print_contents();
  close_printer();
  break;

  case SC_SETUP_VGA_MODE:
  {
   int mode0=*(int *)params,mode;
   vga_modeinfo *i;
   int modes0[]={320,640,800,1024,1200,1600};
   int modes[]=
   {
    G320x200x256,
    G320x200x64K,
    G320x200x16M,
    G320x200x16M32,
    G320x200x32K,
    
    G640x480x256,
    G640x480x64K,
    G640x480x16M,
    G640x480x16M32,
    G640x480x32K,

    G800x600x256,
    G800x600x64K,
    G800x600x16M,
    G800x600x16M32,
    G800x600x32K,

    G1024x768x256,
    G1024x768x64K,
    G1024x768x16M,
    G1024x768x16M32,
    G1024x768x32K,

    G1280x1024x256,
    G1280x1024x64K,
    G1280x1024x16M,
    G1280x1024x16M32,
    G1280x1024x32K,
    
    G1600x1200x256,
    G1600x1200x64K,
    G1600x1200x16M,
    G1600x1200x16M32,
    G1600x1200x32K
   };
   int offset;
   
   #ifdef DG_SCROLLER
    fprintf(stderr,"SC_SETUP_VGA_MODE (%d)\n",mode0);
   #endif
   for (offset=0;offset<sizeof(modes0)/sizeof(*modes0);offset++)
   {
    if (mode0>=modes0[offset]+1&&mode0<=modes0[offset]+5)
    {
     mode=modes[mode0-modes0[offset]-1+5*offset];
     goto ok;
    }
   }
   fprintf(stderr,"scroller: Invalid user graphics-mode code %i\n",mode0);
   return 1; 
   ok:
   if (vga_setmode(mode))
   {
    fprintf(stderr,"Unsuccessful attempt to setup vga mode\n");
    vga_setmode(TEXT);
    return 1;
   }
   if (0<=vga_setlinearaddressing()) vga_linear=1; else vga_linear=0;
   gl_setcontextvga(mode);
   my_graph_mem=vga_getgraphmem();
   i=vga_getmodeinfo(mode);
   if (!i) return 1;
   vga_bytes=i->bytesperpixel;
   xsize=i->width;
   ysize=i->height;
   vga_colors=i->colors;
   vga_linewidth=i->linewidth;
   if (!vga_linear)
   {
    active_page=-1;
    switch_page(0);
   }
   vga_misordered=!!i->flags&RGB_MISORDERED;
   #ifdef DG_SCROLLER
    print_mode_info();
   #endif
   if (setup_color_curves())
   {
    fprintf(stderr,"scroller: can't compute gamma lut.\n");
    return 1;
   }
  }
  break;
  
  case SC_COLOR_CURVES_CHANGED:
  #ifdef DG_SCROLLER
   fprintf(stderr,"SC_COLOR_CURVES_CHANGED\n");
  #endif
  if (setup_color_curves())
  {
   fprintf(stderr,"Scroller: can't SETUP GAMMA\n");
   return 1;
  }
  redraw_screen();
  blit_sl();
  clear_under_sl();
  break;

 }
 return 0;
}

