/* Line with highlighted occurence of searched text
   worm is the number of chars between the two marks. */
static int hstart_line=0,hend_line=0,hstart_char=0,hend_char=0,worm=0;

/* State of searching automatus */
static int a_state=-1;

/* The state of searching automatus finitus */
static int automatus_length;

struct a_entry
{
 unsigned char c;
 int rescue;
};

static struct a_entry *automatus=NULL;
  
/* Init when SC_NEW_SCREEN */  
  hstart_line=0;
  hend_line=0;
  hstart_char=0;
  hend_char=0;
  worm=0;
  a_state=-1;

/* Init when SC_NEW_FONT */
  hstart_line=0;
  hend_line=0;
  hstart_char=0;
  hend_char=0;
  worm=0;
  a_state=-1;

/* Highlights the range hstart_line,hstart_char...hend_line,hend_char. */
void
highlight(void)
{
 int start=max(hstart_line,0);
 int end=min(min(hend_line,sc_b_line+top_lines-1),lines_pos-1);
 int a;

 #ifdef DG_SCROLLER
  fprintf(stderr,"highlight: hstart_line=%i, hstart_char=%i, hend_line=%i,\n"
   "hend_char=%i, worm=%i, a_state=%i.\n",hstart_line, hstart_char, hend_line,
   hend_char, worm, a_state);
 #endif
 for (a=start;a<=end;a++)
  interpret_line(lines[a].ptr,lines[a].number,a-sc_b_line);
}

/* Clears the highlight of the search. */
void
unhighlight(void)
{
 int start=max(hstart_line,sc_b_line);
 int end=min(min(hend_line,sc_b_line+top_lines-1),lines_pos-1);
 int a;
 
 #ifdef DG_SCROLLER
  fprintf(stderr,"Unhighlight.\n");
 #endif
 hstart_line=hend_line=hstart_char=hend_char=worm=0;
 for (a=start;a<=end;a++)
  interpret_line(lines[a].ptr,lines[a].number,a-sc_b_line);
}

/* Performs one step of searching automatus. */
void inline
a_step(unsigned char c)
{
 if (a_state==-1&&(*automatus).c!=c) return;
 if (a_state<automatus_length-1&&automatus[a_state+1].c==c)
 {
  a_state++;
  return;
 }
 retry:
 a_state=automatus[a_state].rescue;
 if (automatus[a_state+1].c==c)
 {
  a_state++;
  return;
 }
 if (a_state==-1) return; /* Initial state */
 goto retry;
}

/* Tries to find next occurence.
   return value: 0 not found
                 1 fatal error
                 2 found
*/
int
find_next(void)
{
 int orig_line=hend_line,orig_char=hend_char;
 int a,b,c,d,e;

 if (!lines_pos) return 0; /* No lines */
 a=hstart_line;
 b=hstart_char;
 c=hend_line;
 d=hend_char;
 e=worm;
 unhighlight();
 hstart_line=a;
 hstart_char=b;
 hend_line=c;
 hend_char=d;
 worm=e;
 do
 {
  #ifdef DG_SCROLLER
   fprintf(stderr,"do: hstart_line=%i, hstart_char=%i, hend_line=%i, hend_char=%i,\n"
    "worm=%i, a_state=%i.\n",hstart_line, hstart_char, hend_line, hend_char,
    worm, a_state);
  #endif
  if (hend_char==lines[hend_line].number)
  {
   a_step(32); /* Space instead of new line. */
  }
  else
  {
   a_step(lines[hend_line].ptr[hend_char]);
  }
  
  /* Now, let's move the highlighted window: */
  if (worm>=automatus_length)
  {
   hstart_char++;
   if (hstart_char>lines[hstart_line].number)
   {
    hstart_char=0;
    hstart_line++;
    if (hstart_line>=lines_pos)
    {
     hstart_line=0;
    }
   }
  }
  else
  {
   worm++;
  }
  hend_char++;
  if (hend_char>lines[hend_line].number)
  {
   hend_char=0;
   hend_line++;
   if (hend_line>=lines_pos)
   {
    hend_line=0;
    {
     #ifdef DG_SCROLLER
      fprintf(stderr,"Search hit BOTTOM.\n");
     #endif
     message("Search hit BOTTOM, continuing at TOP.");
    }
    a_state=-1;
    worm=0;
    hstart_line=0;
    hstart_char=0;
   }
  }
  /* Window moved. Time to test success: */
  if (a_state==automatus_length-1)
  {
   /* Found! */
   highlight();
   return 2;
  }
 }while(orig_line!=hend_line||orig_char!=hend_char);
 return 0; /* Not found */
}

/* Takes a sample and allocates and fills in automatus.
   return value: 1 error
                 0 ok
*/
int
build_automatus(struct block_desc *sample)
{
 int a,rescue,c;
 free(automatus);
 automatus_length=sample->number;
 if (!automatus_length)
 {
  automatus=NULL;
  return 0;
 }
 automatus=(struct a_entry *)malloc(automatus_length*sizeof(struct a_entry));
 if (!automatus)
 {
  fprintf(stderr,"Unable to malloc %i bytes for searching automatus.\n",
   automatus_length*sizeof(struct a_entry));
  return 1;
 }
 automatus[0].c=*sample->ptr;
 automatus[0].rescue=-1;
 for (a=1;a<automatus_length;a++)
 {
  c=sample->ptr[a];
  rescue=automatus[a-1].rescue;
  while(rescue>=0&&automatus[rescue+1].c!=c)
  {
   rescue=automatus[rescue].rescue;
  }
  if (automatus[rescue+1].c==c) rescue++;
  automatus[a].c=c;
  automatus[a].rescue=rescue;
 }
 return 0;
}

  case SC_REGISTER_SAMPLE:
  if (image_mode) break; /* Not allowed in graphics mode */
  #ifdef DG_SCROLLER
   fprintf(stderr,"SC_REGISTER_SAMPLE.\n");
  #endif
  if (build_automatus((struct block_desc *)params))
  {
   fprintf(stderr,"scroller: can't build automatus.\n");
   return 1;
  }
  a_state=-1;
  unhighlight();
  break;
 
  case SC_FIND_NEXT:
  #ifdef DG_SCROLLER
   fprintf(stderr,"SC_FIND_NEXT.\n");
  #endif
  if (image_mode||!automatus) break; /* No search permitted in graphics mode */
  /* Nothing is done when automatus doesn't exist. */
  switch(find_next())
  {
   case 0: /* Not found */
   #ifdef DG_SCROLLER
    fprintf(stderr,"find_next()=0\n");
   #endif
   message("Not found.");
   break;

   case 2: /* Found */
   #ifdef DG_SCROLLER
    fprintf(stderr,"find_next()=2\n");
   #endif
   /* Let's try to skip to the place of occurence */
   {
    int ideal_begin=(hstart_line+hend_line-top_lines)/2;
    move(ideal_begin-sc_b_line);
   }
   break;

   case 1:
   #ifdef DG_SCROLLER
    fprintf(stderr,"find_next()=1\n");
   #endif
   fprintf(stderr,"Unable to find next occurence.\n");
   return 1;
  }
  break;

/* From scroller.c */
void
determine_sh_eh(int *sh, int *eh, int which, int length)
{
 *sh=0;
 if (image_mode||which<0)
 {
  *eh=0;
 }
 else
 {
  if (which>=hstart_line&&which<=hend_line)
  {
   *eh=length;
   if (which==hstart_line) *sh=hstart_char;
   if (which==hend_line) *eh=hend_char;
  }
  else
  {
   *eh=0;
  }
 }
}

/* From scroller.h */  
/* Registers a search keyword. Subsequent calls of SC_FIND_NEXT will find it
   In the time when you register a sample, no searching action is performed.
   The searching pointer (and the highligted window) is set to start and
   begin at the very begin of the text.
   When not in text mode, everything is ok.
   parameters: pointer to a null-terminated string defining the sample
*/
SC_REGISTER_SAMPLE,

/* Finds next occurence. Maybe issue the message "Not found." or
   "Search hit BOTTOM, continuing at TOP."
   parameters: nothing
*/
SC_FIND_NEXT,

/* From watcher.c from SCANCODE_ENTER */
     case 2:
     #ifdef DEBUG_SEARCH
      printf("SCANCODE_ENTER && input_state==2.\n");
     #endif
     input_state=0;
     string_from_input(input_search);
     {
      struct block_desc a;
      a.ptr=input_buffer;
      a.number=input_buffer_length;
      if (scroller(SC_REGISTER_SAMPLE,&a))
      {
       fprintf(stderr,"watcher: can't register a sample calling scroller.\n");
       return 1;
      }
     }
     if (scroller(SC_FIND_NEXT,NULL))
     {
      fprintf(stderr,"watcher: unable to find next occurence calling"
       " scroller.\n");
      return 1;
     }
     break;
   
/* Fro watcher.c test for key 'n' */   
   else
   {
    if (scroller(SC_FIND_NEXT,NULL))
    {
     fprintf(stderr,"watcher: unable to find next occurence calligng"
      " scroller.\n");
     return 1;
    }
   }
   
/* From test for SCANCODE_SLASH */
   else
   {
    string_to_input(input_search);
    draw_input_buffer();
    input_state=2; /* Search keyword */
   }

