/* Svgalib Web Browser by Clocksoft */
/* (c) 1999 Clocksoft */
/* Watcher -- watches for incoming data and keyhits */

#define OTHK
//#define WATCHDOG
/* When no keyboard scan is performed during 10 seconds, the program ends. */

#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <vga.h>
#include <vgakeyboard.h>
#include <signal.h>
#include <sys/ioctl.h>
#ifndef OTHK
#include <linux/kd.h>
#endif
#include <sys/vt.h>
#include <sys/wait.h>

#include "gvars.h"
#include "line_breaker.h"
#include "scroller.h"
#include "font_loader.h"
#include "watcher.h"
#include "image_parser.h"

//#define DEBUG_WATCHER
//#define DEBUG_VT
//#define DEBUG_FILE

#define FILE_BUFFER_SIZE 4096
#define min(x,y) ((x)<(y)?(x):(y))

#ifndef OTHK
extern int __svgalib_tty_fd;
#endif

/********************* Global variables for watcher ***************************/

/* Buffer for reading data from the file */
unsigned char fbuf[FILE_BUFFER_SIZE];

/* Buffer for input like search keyword or "enter URL" and so on */
enum {max_ib=256};
static unsigned char input_buffer[max_ib+1];
static unsigned char input_load[max_ib+1];
static unsigned char input_search[max_ib+1];

/* How many chars in input buffer are valid */
static int input_buffer_length=0;

/* A byte offset in buffer of displayed area. */
/* It's forbidden to have input_buffer_display_start>input_buffer_length */
static int input_buffer_display_start=0;

/* Tells how wide is the visible part of input_buffer. In pixels.*/
static int input_buffer_size=0;

/* Tells what is being input: 0 nothing
                              1 file name to load
                              2 search keyword
*/
static int input_state;

#ifndef OTHK
/* The keyboard mode when the prog was run */
static int old_keyboard_mode;
#endif

/* Shift flag */
int leftshift,rightshift,leftalt,rightalt,rightctrl,leftctrl,capslock;

/* Packing up/downs/lefts/rights/pgups/pgdns */
enum{M_N,M_U,M_D,M_L,M_R,M_PU,M_PD};
int merge=M_N;
unsigned merge_count=1;

#ifndef OTHK
/* Old termios from keyboard */
static struct termios old_termios;
#endif

#ifdef OTHK
static int keyboard_fd;
#endif

static pid_t lynx_pid;
pid_t lpr_pid;
static volatile int lynx_runs;
volatile int lpr_runs;

/* The *ptr must contain malloced pointer or NULL */
int
copy_string(unsigned char **ptr, const char *string)
{
 size_t len;
 char *p;
 len=strlen(string);
 len++;
 free(*ptr);
 p=(char *)malloc(len);
 if (!p)
 {
  fprintf(stderr,"new_string: out of memory.\n");
  return 1;
 }
 memcpy(p,string,len);
 *ptr=p;
 return 0;
}

int derr(unsigned char *txt)
/* Send the error message the line breaker and then flushes broken line into
   scroller. */
/* 1=error 0=success */
{
 struct block_desc block;
 #ifdef DEBUG_WATCHER
  printf("I got into derr.\n");
 #endif
 block.ptr=txt;
 block.number=strlen(txt);
 if (line_breaker(LB_TYPESET_BLOCK,&block))
 {fprintf(stderr,"Unable to typeset block.\n");return 1;}
 if (line_breaker(LB_FLUSH_LINE,NULL))
 {fprintf(stderr,"Unable to end reloading.\n");return 1;}
 return 0;
}

/* Restores the original keyboard setting. */
int
reset_keyboard(void)
{
 #ifdef OTHK
 keyboard_close();
 #else
 if (ioctl(__svgalib_tty_fd,KDSKBMODE,old_keyboard_mode))
 {
  fprintf(stderr,"Bad news: I didn't manage to return the keyboard to the"
   " original mode.\n");
  return 1;
 }
 tcsetattr(__svgalib_tty_fd,TCSAFLUSH,&old_termios);
 #endif
 return 0;
}

void sighandler(int signal)
{
 #ifdef WATCHDOG
 if (signal==SIGALRM)
 {
  fprintf(stderr,"Watchdog: something is wrong.\n");
 }
 else
 {
 #endif 
  psignal(signal,"wb0");
 #ifdef WATCHDOG
 }
 #endif
 reset_keyboard();
 vga_setmode(TEXT);
 fflush(stdout);
 fflush(stderr);
 _exit(1);
}

void
close_source(void)
{
 if (lynx_runs)
 {
  signal(SIGCHLD,SIG_DFL);
  if (lynx_runs)
  {
   kill(lynx_pid,SIGKILL);
   wait(NULL);
   lynx_runs=0;
  }
 }
 close(source);
 source=-1;
}

/* Takes up up to (buffer size) bytes and flushes into line breaker. 
   If the content_type is not WAI_TEXT, then the data don't go into line breaker
   but into image_parser.
*/
int
pour_block_into_lb(void)
{
 unsigned char *p; /* Points at first byte still not processed. */
 ssize_t count,len; /* count=how much still not processed. */
 struct block_desc block;
 void *match;
 int eof; /* Flag indicating whether read reported eof or not. */
 
 #ifdef DEBUG_WATCHER
  printf("I got into pour_block_into_lb.\n");
 #endif
 if (source<0)
 {
  fprintf(stderr,"Internal error: pour_block_into_lb called"
   " with source=%i.\n",source);
  return 1;
 }
 count=read(source,fbuf,FILE_BUFFER_SIZE);
 #ifdef DEBUG_WATCHER
  printf("%i bytes read by read.\n",count);
 #endif
 eof=(count<=0&&!lynx_runs);
 p=fbuf;
 if (!eof)
 {
  /* No eof */
  if (count<=0)
  {
   return 0;
  }
  if (content_type==WAI_TEXT||content_type==WAI_HTML) /* Crippled! */
  {
   while(count>0)
   {match=memchr(p,'\n',count);
    if (match)
    {/* newline found at *match */
     len=(unsigned char *)match-p;
     block.number=len;
     block.ptr=p;
     #ifdef DEBUG_WATCHER
      printf("Going to typeset %i bytes and one linefeed.\n",len);
     #endif
     if (line_breaker(LB_TYPESET_BLOCK,&block))
     {fprintf(stderr,"Unable to typeset block right before newline. Length="
       "%i\n",len);
      return 1;
     }
     len++; /* Let's take into account also the '\n'. */
     p+=len;
     count-=len;
     if (line_breaker(LB_FLUSH_LINE,NULL))
     {fprintf(stderr,"Unable to flush line for a newline.\n");
      return 1;
     }
    }
    else
    {/* newline not found */
     block.number=count;
     block.ptr=p;
     #ifdef DEBUG_WATCHER
      printf("Going to typeset %i bytes.\n",count);
     #endif
     if (line_breaker(LB_TYPESET_BLOCK,&block))
     {fprintf(stderr,"Unable to typeset ordinary block %i long.\n",count);
      return 1;
     }
     count=0; /* We have made it all at once */
    }
   }
  }
  else
  {
   block.ptr=fbuf;
   block.number=count;
   #ifdef DEBUG_WATCHER
    printf("Sending read data into image parser.\n");
   #endif
   switch(image_parser(IP_PARSE_BLOCK,&block))
   {
    case 1:
    /* Fatal error */
    fprintf(stderr,"Unable to parse a block of image.\n");
    return 1; /* Fatal error */
    
    case 2:
    #ifdef DEBUG_WATCHER
     printf("watcher: stop sending data received from image parser.\n");
    #endif
    /* Stop sending data */
    close_source();
    return 0;
   }
  }
 }
 else
 {
  /* There is an eof */
  if (content_type==WAI_TEXT||content_type==WAI_HTML)
  {
   if (line_breaker(LB_FLUSH_LINE,NULL))
   {fprintf(stderr,"Unable to flush line at end of file.\n");
    return 1;
   }
  }
  else
  {
   /* image_parser needs an end-of-file indication. */
   switch (image_parser(IP_EOF,NULL))
   {
    case 2:
    #ifdef DEBUG_WATCHER
     printf("watcher: stop sending data received from image parser"
      " after an EOF.\n");
    #endif
    /* Stop sending data */
    close_source();
    return 0;
    
    case 1:
    fprintf(stderr,"watcher: unable to send EOF to image_parser.\n");
    return 1;
   }
  }
  close_source();
 }
 return 0;
}

/* Prints out a message telling something about encoding being loaded. */
int
announce_encoding(void)
{struct block_desc b;
 const unsigned char txt0[]="Encoding ";
 const unsigned char txt1[]=". ";
 unsigned char *txt; 
 int txtlen;
 
 txtlen=strlen(txt0)+strlen(txt1)+strlen(actual_encoding->name);
 txt=(unsigned char *)malloc(txtlen+1); /* Ending \0 !!! */
 if (!txt)
 {
  fprintf(stderr,"Unable to allocate small buffer %i bytes long for "
   "building a message about encoding change.\n",txtlen+1);
  return 1;
 }
 strcpy(txt,txt0);
 strcat(txt,actual_encoding->name);
 strcat(txt,txt1);
 
 b.ptr=txt;
 b.number=txtlen;
 if (scroller(SC_DISPLAY_LINE,&b))
 {
  fprintf(stderr,"Unable to announce \"%s\".\n",txt);
  return 1;
 }
 free(txt);
 return 0; 
}

/* Clears the onscreen space for input buffer */
int
draw_black_input_buffer(void)
{
 struct block_desc black;
 black.ptr=NULL;
 black.number=0;
 if (scroller(SC_DISPLAY_LINE,&black))
 {
  fprintf(stderr,"Can't display empty status line.\n");
  return 1;
 }
 return 0;
}

int
draw_input_buffer(void)
{ 
 struct block_desc b;
 
 b.ptr=input_buffer+input_buffer_display_start;
 b.number=input_buffer_length-input_buffer_display_start;
 if (scroller(SC_DISPLAY_LINE,&b))
 {
  fprintf(stderr,"Unable to display input line calling scroller.\n");
  return 1;
 }
 return 0;
}

/* The result is in "input buffer" and the length is in "input_buffer_length".
   256 delete
   257 kill whole line (^N)
   258 nothing
   return value: 0 ok
                 1 error
*/
int
input(unsigned c)
{
 
 switch(c)
 {
  case 256:
  /* Deleting last char in input_buffer */
  if (!input_buffer_length) return 0; /* Nothing to delete */
  input_buffer_length--;
  input_buffer_size-=font_widths[input_buffer[input_buffer_length]];
  while(input_buffer_display_start>0&&input_buffer_size
   +font_widths[input_buffer[input_buffer_display_start-1]]<=xsize)
  {
   /* Shifting line right in order to display as much as possible */
   input_buffer_display_start--;
   input_buffer_size
    +=font_widths[input_buffer[input_buffer_display_start]];
  }
  if (draw_input_buffer())
  {
   fprintf(stderr,"Unable to redraw input_buffer after backspace was "
    "pressed.\n");
   return 1;
  }
  break;
  
  case 257:
  input_buffer_length=0;
  input_buffer_size=0;
  input_buffer_display_start=0;
  if (draw_input_buffer())
  {
   fprintf(stderr,"Unable to redraw input_buffer after kill was "
    "pressed.\n");
   return 1;
  }
  break;
  
  case 258:
  break;
  
  default:
  if (input_buffer_length==max_ib) return 0; /* input_buffer full */
  input_buffer[input_buffer_length]=c;
  input_buffer_length++;
  input_buffer_size+=font_widths[c];
  while(input_buffer_size>xsize)
  { 
   input_buffer_size-=font_widths[input_buffer[input_buffer_display_start]];
   input_buffer_display_start++;
  }
  if (draw_input_buffer())
  {
   fprintf(stderr,"Unable to draw input_buffer.\n");
   return 1;
  }
  break;
 }
 return 0;
}

/* Takes "filename" global variable and determines using the filename's end
   the type of the file. Later, this should probably use the content-type.
   Return value: 0 ok
                 1 fatal error
*/
int
type_from_fname(void)
{
 int l;

 content_type=WAI_TEXT;
 if (!*filename) return 0; /* No filename, text document :) */
 l=strlen(filename);
 if (l<3)
 {
  return 0;
 }
 if (!strncasecmp(filename+l-3,"gif",3))
 {
  content_type=WAI_GIF;
  return 0;
 }
 if (!strncasecmp(filename+l-3,"pcx",3))
 {
  content_type=WAI_PCX;
  return 0;
 }
 if (!strncasecmp(filename+l-3,"png",3))
 {
  content_type=WAI_PNG;
  return 0;
 }
 if (!strncasecmp(filename+l-3,"jpg",3)||!strncasecmp(filename+l-4,"jpeg",4))
 {
  content_type=WAI_JPEG;
  return 0;
 }
 if (!(strncasecmp(filename+l-3,"htm",3)&&strncasecmp(filename+l-4,"html",
  4)))
 {
  content_type=WAI_HTML;
  return 0;
 }
 return 0;
}

/* Flush cached mergeable keystrokes */
int
flush_merge(void)
{
 switch(merge)
 {
  case M_U:
  if (scroller(SC_LINE_UP,&merge_count))
  {
   fprintf(stderr,"Unable to scroll %u lines up.\n",merge_count);
   return 1;
  }
  break;

  case M_D:
  if (scroller(SC_LINE_DOWN,&merge_count))
  {
   fprintf(stderr,"Unable to scroll %u lines down.\n",merge_count);
   return 1;
  }
  break;

  case M_PU:
  if (scroller(SC_PAGE_UP,&merge_count))
  {
   fprintf(stderr,"Unable to scroll %u pages up.\n",merge_count);
   return 1;
  }
  break;

  case M_PD:
  if (scroller(SC_PAGE_DOWN,&merge_count))
  {
   fprintf(stderr,"Unable to scroll %u pages down.\n",merge_count);
   return 1;
  }
  break;

  case M_L:
  if (scroller(SC_LEFT,&merge_count))
  {
   fprintf(stderr,"Unable to scroll %u times left.\n",merge_count);
   return 1;
  }
  break;

  case M_R:
  if (scroller(SC_RIGHT,&merge_count))
  {
   fprintf(stderr,"Unable to scroll %u times right.\n",merge_count);
   return 1;
  }
  break;
 }
 merge=M_N;
 return 0;
}

/* Process a mergeable keystroke */
int
inc_merge(int what)
{
 if (what==merge) 
 {
  merge_count++;
  return 0;
 }
 if (flush_merge())
 {
  return 1;
 }
 merge=what;
 merge_count=1;
 return 0;
}

void
child_handler(int a)
{
 pid_t pid;
 pid=wait(NULL);
 if (lynx_runs&&pid==lynx_pid)
 {
  lynx_runs=0;
 }
 if (lpr_runs&&pid==lpr_pid)
 {
  fprintf(stderr,"lpr exited\n");
  lpr_runs=0;
 }
 signal(SIGCHLD,child_handler);
}

/* Tells whether the filename points to local file or to remote. */
int
is_local()
{
 return (strncmp("http://",filename,7)&&strncmp("ftp://",filename,6));
}

/* Returns 0:ok
           1:error */
int
open_source(void)
{
 if (is_local()&&content_type!=WAI_HTML)
 {
  source=open(filename,O_RDONLY);
  if (source<0&&derr("Unable to open that local file."))
  {
   fprintf(stderr,"Unable to issue unability to open local file.\n");
   return 1;
  }
  if (source>=0)
  {
   fcntl(source,F_SETFL,O_NONBLOCK);
  }
  else
  {
   source=-1;
  }
  return 0;
 }
 else
 {
  int fds[2];
  if (pipe(fds))
  {
   fprintf(stderr,"Unable to create a pipe.\n");
   return 1;
  }
  signal(SIGCHLD,child_handler);
  lynx_pid=fork();
  if (lynx_pid)
  {
   /* I am the parent */
   lynx_runs=1;
   source=fds[0];
   close(fds[1]);
   if (fcntl(fds[0],F_SETFL,fcntl(fds[0],F_GETFL)|O_NONBLOCK))
   {
    fprintf(stderr,"Unable to fcntl.\n");
    close_source();
    return 1;
   }
  }
  else
  {
   /* I am the child */
   close(fds[0]);
   if (-1==dup2(fds[1],1))
   {
    fprintf(stderr,"Child: unable to dup2.\n");
    _exit(-1);
   }
   close(fds[1]);
   if (content_type==WAI_HTML||content_type==WAI_TEXT)
   { 
    execlp("lynx","lynx","-dump","-raw","-cfg=/etc/wb0/wb0.lynx.cfg",filename,NULL);
   }
   else 
   { 
    execlp("lynx","lynx","-source","-cfg=/etc/wb0/wb0.lynx.cfg",filename,NULL);
   }
   fprintf(stderr,"Unable to execlp lynx.\n");
   _exit(-1); /* Error */
  }
 } 
 return 0;
}

/* Takes the filename and reopens the file. Includes partial reinit of the
   whole browser.
*/
int
reopen_file(void)
{
 #ifdef DEBUG_FILE
  printf("Getting into reopen_file. Filename=%s\n",filename);
 #endif
 /* Let's determine the type of the file */
 if (source>=0)
  close_source();
 #ifdef DEBUG_FILE
  printf("source closed.\n");
 #endif
 if (type_from_fname())
 {
  fprintf(stderr,"Unable to determine type of the file from the filename.\n");
  return 1;
 }
 #ifdef DEBUG_FILE
  printf("type_from_fname OK.\n");
 #endif
 if ((content_type!=WAI_TEXT&&content_type!=WAI_HTML)||calibrate)
 {
  int type;
  
  switch (content_type)
  {
   case WAI_GIF: type=0; break;
   case WAI_PCX: type=1; break;
   case WAI_PNG: type=3; break;
   default: type=2; break; /* jpeg */
  }
  if (image_parser(IP_NEW_SCREEN,&type))
  {
   fprintf(stderr,"watcher: Unable to reinitialize image parser.\n");
   return 1;
  }
 }
 #ifdef DEBUG_FILE
  printf("content_type OK.\n");
 #endif
 {
  int pal_mode=(content_type!=WAI_TEXT&&content_type!=WAI_HTML)||calibrate;
  if (line_breaker(LB_NEW_SCREEN,&pal_mode))
  {
   fprintf(stderr,"watcher: unable to reinit line breaker after"
    " renewing screen or font.\n");
   return 1;
  }
  pal_mode=0;
 }
 #ifdef DEBUG_FILE
  printf("Line_breaker reinit ok.\n");
 #endif
 if (calibrate)
 {
  if (image_parser(IP_CALIBRATE,NULL))
  {
   fprintf(stderr,"Unable to calibrate.\n");
   return 1;
  }
  calibrate=0;
 }
 else if (!*filename)
 {
  if (derr("No filename specified."))
  {
   fprintf(stderr,"watcher: unable to issue error message \"No filename "
    "specified.\".\n");
   return 1;
  }
 }
 else
 {
  if (open_source()&&derr("Unable to open source."))
  {
   fprintf(stderr,"watcher: unable to issue message \"Unable to "
    "open source.\".\n");
   return 1;
  }
 }
 #ifdef DEBUG_WATCHER
  printf("Leaving reopen_file.\n");
  printf("source=%i, filename=%s.\n",source, filename);
 #endif
 return 0;
}

/* If s is null, it only recomputes the input_buffer counters.
   If the string is longer than max_ib, it is truncated. */
void
string_to_input(unsigned char *s)
{
 int tmp;

 if (s)
 {
  input_buffer_length=min(max_ib,strlen(s));
  memcpy(input_buffer,s,input_buffer_length+1);
 }
 input_buffer_display_start=input_buffer_length;
 input_buffer_size=0;
 while(input_buffer_display_start)
 {
  tmp=input_buffer_size+font_widths[input_buffer[
   input_buffer_display_start-1]];
  if (tmp>xsize) break;
  input_buffer_size=tmp;
  input_buffer_display_start--;
 }
}

/* Makes a string from input and stores it into "dest".
   There must be at least max_ib+1 bytes in the "dest".
*/
void
string_from_input(unsigned char *dest)
{
 input_buffer[input_buffer_length]=0;
 memcpy(dest,input_buffer,max_ib+1);
}

/* REVerse Encoding Table lookup
   return value: 258 not found
                 0...255 found
*/
int
revet(int what)
{
 int c;
 
 if (what<0||what>=65536) return 258;
 for (c=0;c<256;c++)
  if (encoding[c]==what) return c;
 return 258;
}

/* Tests the keyboard and does the actions. */
int
process_keys(void)
{
 unsigned char k;

 #ifdef WATCHDOG
  alarm(10);
 #endif
 #ifdef OTHK
 while(1==read(keyboard_fd,&k,1))
 #else
 while(1==read(__svgalib_tty_fd,&k,1))
 #endif
 {
  switch(k)
  {
   case SCANCODE_CURSORBLOCKRIGHT:
   if (inc_merge(M_R))
   {
    return 1;
   }
   continue;
   
   case SCANCODE_CURSORBLOCKLEFT:
   if (inc_merge(M_L))
   {
    return 1;
   }
   continue;
   
   case SCANCODE_CURSORBLOCKUP:
   if (inc_merge(M_U))
   {
    return 1;
   }
   continue;
   
   case SCANCODE_CURSORBLOCKDOWN:
   if (inc_merge(M_D))
   {
    return 1;
   }
   continue;
   
   case SCANCODE_PAGEUP:
   if (inc_merge(M_PU))
   {
    return 1;
   }
   continue;
   
   case SCANCODE_PAGEDOWN:
   if (inc_merge(M_PD))
   {
    return 1;
   }
   continue;
  
   case SCANCODE_LEFTALT:
   leftalt=1;
   break;
   
   case SCANCODE_RIGHTALT:
   rightalt=1;
   break;
   
   case SCANCODE_LEFTSHIFT:
   leftshift=1;
   break;
   
   case SCANCODE_RIGHTSHIFT:
   rightshift=1;
   break;
   
   case SCANCODE_LEFTCONTROL:
   leftctrl=1;
   break;
   
   case SCANCODE_RIGHTCONTROL:
   rightctrl=1;
   break;

   case SCANCODE_CAPSLOCK:
   capslock=1;
   break;
   
   case 0x80|SCANCODE_LEFTALT:
   leftalt=0;
   break;
   
   case 0x80|SCANCODE_RIGHTALT:
   rightalt=0;
   break;
   
   case 0x80|SCANCODE_LEFTSHIFT:
   leftshift=0;
   break;
   
   case 0x80|SCANCODE_RIGHTSHIFT:
   rightshift=0;
   break;
   
   case 0x80|SCANCODE_LEFTCONTROL:
   leftctrl=0;
   break;
   
   case 0x80|SCANCODE_RIGHTCONTROL:
   rightctrl=0;
   break;
   
   case 0x80|SCANCODE_CAPSLOCK:
   capslock=0;
   break;
   
   case 59:
   case 60:
   case 61:
   case 62:
   case 63:
   case 64:
   case 65:
   case 66:
   case 67:
   case 68:
   case 87:
   case 88:
   if (leftalt||rightalt)
   {
    /* alt-f1 ... alt-f10, alt-f11 ... alt-f12 */
    int vt;
    struct vt_stat vts;
    
    if (k>=87)
    {vt=11+k-87;
    }
    else
    {vt=1+k-59;
    }
    #ifdef OTHK
    ioctl(keyboard_fd,VT_GETSTATE,&vts);
    #else
    ioctl(__svgalib_tty_fd,VT_GETSTATE,&vts);
    #endif
    if (vt!=vts.v_active)
    { 
     rightshift=leftshift=rightalt=leftalt=leftctrl=rightctrl=capslock=0;
     #ifdef DEBUG_VT
     printf("Trying to switch VT.\n");
     #endif
     #ifdef OTHK
     ioctl(keyboard_fd,VT_ACTIVATE,vt);
     #else
     ioctl(__svgalib_tty_fd,VT_ACTIVATE,vt);
     #endif
    }
   }
   break;

   case SCANCODE_TAB:
   {/* switch encoding */
    if (leftshift||rightshift)
    {
     struct enc_member *before=actual_encoding;
     if (before==enc_list)
     {
      before=NULL;
     }
     while(actual_encoding->next!=before)
     {
      actual_encoding=actual_encoding->next;
      if (!actual_encoding) actual_encoding=enc_list;
     }
     /* Backward switch */
    }
    else
    {
     /* Forward switch */
     if (!(actual_encoding=actual_encoding->next))
      actual_encoding=enc_list;
    }
    if (announce_encoding())
    {
     fprintf(stderr,"Unable to announce encoding.\n");
     return 1;
    }
    if (font_loader())
    {fprintf(stderr,"Unable to reload font after encoding change.\n");
     return 1;
    }
    if (line_breaker(LB_NEW_FONT,NULL))
    {fprintf(stderr,"Unable to reset line breaker after font change.\n");
     return 1;
    }
    if (reopen_file())
    {
     fprintf(stderr,"Unable to reopen file after encoding switch.\n");
     return 1;
    }
   }
   break;

   #define X(a,b,c) case a: if (input_state){if(rightshift||leftshift)input(b);else input(c);} break;
   #define Z(a,b,c,d,e) case a:if(input_state){if(rightshift||leftshift){if(capslock)input(revet(e));else input(b);}else {if (capslock)input(revet(d));else input(c);}}break;
   #define Y(a,b) case a:if(input_state)input(b);break;
   Z(SCANCODE_A,'A','a',0xe1,0xc1)
   Z(SCANCODE_B,'B','b',-1,-1)
   Z(SCANCODE_C,'C','c',0x10d,0x10c)
   Z(SCANCODE_D,'D','d',0x10f,0x10e)
   Z(SCANCODE_E,'E','e',0xe9,0xc9)
   Z(SCANCODE_F,'F','f',0xeb,0xcb)
   Z(SCANCODE_G,'G','g',-1,-1)
   Z(SCANCODE_H,'H','h',0xfa,0xda)
   Z(SCANCODE_I,'I','i',0xed,0xcd)
   Z(SCANCODE_J,'J','j',0xfc,0xdc)
   Z(SCANCODE_K,'K','k',0xf4,0xd4)
   case SCANCODE_L:
   if (input_state)
   {
    if (rightshift||leftshift)
    {
     if (capslock)
     {
      input(revet(0x13d));
     }
     else
     {
      input('L');
     }
    }
    else
    { 
     if (capslock)
     {
      input(revet(0x13e));
     }
     else
     {
      input('l');
     }
    }
   }
   else
   {
    string_to_input(input_load);
    draw_input_buffer();
    input_state=1;
   }
   break;
   
   Z(SCANCODE_M,'M','m',0x13a,0x139)
   case SCANCODE_N:
   if (input_state)
   {
    if (leftctrl||rightctrl)
    {
     input(257);
    }
    else if (leftshift||rightshift)
    {
     if (capslock)
     {
      input(revet(0x147));
     }
     else
     {
      input('N');
     }
    }
    else
    {
     if (capslock)
     {
      input(revet(0x148));
     }
     else
     {
      input('n');
     }
    }
   }
   break;

   Z(SCANCODE_O,'O','o',0xf3,0xd3)
   case SCANCODE_P:
   if (input_state)
   {
    if (leftshift||rightshift)
    {
     if (capslock)
     {
      input(revet(0xd6));
     }
     else
     {
      input('P');
     }
    }
    else
    {
     if (capslock)
     {
      input(revet(0xf6));
     }
     else
     {
      input('p');
     }
    }
   }
   else
   {
    if (scroller(SC_PRINT,NULL))
    {
     fprintf(stderr,"watcher.c: unable to print hardcopy\n");
     return 1;
    }
   }
   break;

   case SCANCODE_Q:
   case SCANCODE_ESCAPE:
   if (!input_state)
   {
    #ifdef DEBUG_WATCHER
     printf("Quitting due to pressed 'Q' or ESC.\n");
    #endif
    return 2; /* Quit the program. */
   }
   if (k==SCANCODE_Q)
   {
    if (rightshift||leftshift)
    {
     if (capslock)
     {
      input(revet(0xc4));
     }
     else
     {
      input('Q');
     }
    }
    else
    { 
     if (capslock)
     {
      input(revet(0xe4));
     }
     else
     {
      input('q');
     }
    }
   }
   else 
   {
    switch(input_state)
    {
     case 1:
     string_from_input(input_load);
     break;

     case 2:
     string_from_input(input_search);
     break;
    }
    if (draw_black_input_buffer())
    {
     fprintf(stderr,"Unable to display black input buffer.\n");
     return 1;
    }
    input_state=0;
   }
   break;
   
   Z(SCANCODE_R,'R','r',0x159,0x158)
   Z(SCANCODE_S,'S','s',0x161,0x160)
   Z(SCANCODE_T,'T','t',0x165,0x164)
   case SCANCODE_U:
   if (input_state)
   {
    if (leftctrl||rightctrl)
    {
     input(257);
    }
    else if (leftshift||rightshift)
    {
     if (capslock)
     {
      input(revet(0x16e));
     }
     else
     {
      input('U');
     }
    }
    else
    {
     if (capslock)
     {
      input(revet(0x16f));
     }
     else
     {
      input('u');
     }
    }
   }
   break;

   Z(SCANCODE_V,'V','v',0xe7,0xc7)
   Z(SCANCODE_W,'W','w',0x11b,0x11a)
   Z(SCANCODE_X,'X','x',-1,-1)
   Z(SCANCODE_Y,'Y','y',0xfd,0xdd)
   Z(SCANCODE_Z,'Z','z',0x17e,0x17d)
   X(SCANCODE_MINUS,'_','-')
   X(SCANCODE_EQUAL,'+','=')
   X(SCANCODE_BACKSLASH,'|','\\')
   X(SCANCODE_PERIOD,'>','.')
   X(SCANCODE_COMMA,'<',',')
   case SCANCODE_SLASH:
   if (input_state)
   {
    if (leftshift||rightshift)
    {
     input('?');
    }
    else
    {
     input('/');
    }
   }
   break;
   
   X(SCANCODE_SEMICOLON,':',';')
   X(SCANCODE_APOSTROPHE,'"','\'')
   X(SCANCODE_BRACKET_LEFT,'{','[')
   X(SCANCODE_BRACKET_RIGHT,'}',']')
   X(SCANCODE_0,')','0')
   X(SCANCODE_1,'!','1')
   X(SCANCODE_2,'@','2')
   X(SCANCODE_3,'#','3')
   X(SCANCODE_4,'$','4')
   X(SCANCODE_5,'%','5')
   X(SCANCODE_6,'^','6')
   X(SCANCODE_7,'&','7')
   X(SCANCODE_8,'*','8')
   X(SCANCODE_9,'(','9')
   Y(SCANCODE_KEYPAD0,'0');
   Y(SCANCODE_KEYPAD1,'1');
   Y(SCANCODE_KEYPAD2,'2');
   Y(SCANCODE_KEYPAD3,'3');
   Y(SCANCODE_KEYPAD4,'4');
   Y(SCANCODE_KEYPAD5,'5');
   Y(SCANCODE_KEYPAD6,'6');
   Y(SCANCODE_KEYPAD7,'7');
   Y(SCANCODE_KEYPAD8,'8');
   Y(SCANCODE_KEYPAD9,'9');
   X(SCANCODE_SPACE,' ',' ');
   X(SCANCODE_GRAVE,'~','`');

   case SCANCODE_BACKSPACE:
   if (input_state&&input(256)) /* Delete */
   {
    fprintf(stderr,"Unable to do input backspace.\n");
    return 1;
   }
   break;

   case SCANCODE_ENTER:
   #ifdef DEBUG_FILE
    printf("case SCANCODE_ENTER:\n");
   #endif
   if (input_state)
   {
    switch(input_state)
    {
     case 1: 
     #ifdef DEBUG_FILE
      printf("SCANCODE_ENTER && input_state==1.\n");
     #endif
     input_state=0;
     string_from_input(input_load);
     copy_string(&filename,input_load);
     if (reopen_file())
     {
      fprintf(stderr,"Unable to reopen file.\n");
      return 1;
     }
     #ifdef DEBUG_FILE
      printf("reopen OK.\n");
     #endif
     break;
    }
   }
   break;

   case SCANCODE_HOME:
   if (scroller(SC_HOME,NULL))
   {fprintf(stderr,"Unable to scroll to the top of the page.\n");
    return 1;
   }
   break;

   case SCANCODE_END:
   if (scroller(SC_END,NULL))
   {fprintf(stderr,"Unable to scroll to the bottom of the page.\n");
    return 1;
   }
   break;
  }
  if (flush_merge())
  {
   return 1;
  }
 }
 if (flush_merge())
 {
  return 1;
 }
 #ifdef DEBUG_WATCHER
  printf("Leaving process_keys.\n");
 #endif
 usleep(100);
 return 0;
}

int watcher1(void)
/* Scans keyboard (raw mode) and opens file handle (source) for incoming
   things.
*/
{ 
 
 #ifdef DEBUG_WATCHER
  printf("I got into watcher1.\n");
 #endif
 if (font_loader())
 {fprintf(stderr,"Unable to load font.\n");
  return 1;
 }
 string_to_input(filename);
 string_from_input(input_load);
 if (reopen_file())
 {
  fprintf(stderr,"Unable to reopen file.\n");
  return 1;
 }
 if (announce_encoding())
 {
  fprintf(stderr,"Unable to announce encoding.\n");
  return 1;
 }
 while(1)
 {
  if (source>=0)
  {/* If the source is not closed */
   #ifdef DEBUG_WATCHER
    printf("watcher1: reading block.\n");
   #endif
   if (pour_block_into_lb())
   /* pour_block_into_lb can also close the source (and must then set */
   /* source=-1) */
   {fprintf(stderr,"Unable to pour incoming data into line breaker.\n");
    return 1;
   }
  }
 
  #ifdef DEBUG_WATCHER
   printf("source before process_keys=%i.\n",source);
  #endif
  switch (process_keys())
  {
   case 1:
   fprintf(stderr,"Unable to process keys.\n");
   return 1;

   case 2:
   return 0; /* A time to quit. */
  }
  #ifdef DEBUG_WATCHER
   printf("source after process_keys=%i.\n",source);
  #endif
 }
}

/* Saves the mode and sets up raw mode. */
int
set_keyboard(void)
{
 int sigs[]={SIGINT,SIGQUIT,SIGTERM,SIGSEGV,SIGHUP,SIGILL,SIGABRT,SIGFPE,
  SIGBUS,SIGPWR,SIGSTKFLT,SIGALRM};
 int *p;
 int i;
 
 #ifdef OTHK
 keyboard_fd=keyboard_init_return_fd();
 if (keyboard_fd<0)
 {
  fprintf(stderr,"Unable to init_keyboard().\n");
  return 1;
 }
 #else
 if (ioctl(__svgalib_tty_fd,KDGKBMODE,&old_keyboard_mode))
 {
  fprintf(stderr,"Unable to get keyboard mode.\n");
  return 1;
 }
 tcgetattr(__svgalib_tty_fd,&old_termios);
 {
  struct termios t=old_termios;
  t.c_lflag&=~(ICANON|ECHO|ISIG);
  t.c_iflag&=~(ISTRIP|IGNCR|ICRNL|INLCR|IXOFF|IXON);
  t.c_cc[VMIN]=0;
  t.c_cc[VTIME]=0;
  tcsetattr(__svgalib_tty_fd,TCSAFLUSH,&t);
 }
 {
  int f;
  f=fcntl(__svgalib_tty_fd,F_GETFL);
  f|=O_NONBLOCK;
  if (fcntl(__svgalib_tty_fd,F_SETFL,f))
  {
   fprintf(stderr,"Unable to fcntl keyboard to nonblocking mode.\n");
   return 1;
  }
 }
 if (ioctl(__svgalib_tty_fd,KDSKBMODE,K_MEDIUMRAW))
 {
  fprintf(stderr,"Unable to set keyboard to medium-raw mode.\n");
  return 1;
 }
 #endif
 for (i=sizeof(sigs)/sizeof(int),p=sigs;i;i--)
  if (SIG_ERR==signal(*p++,sighandler))
  {fprintf(stderr,"Unable to hang up signal handler.\n");
   return 1;
  }
 signal(SIGPIPE,SIG_IGN);
 #ifdef WATCHDOG 
  alarm(10);
 #endif
 return 0;
}

int watcher(void)
{int status;
 if (set_keyboard())
 {
  return 1;
 }
 status=watcher1();
 reset_keyboard();
 return status;
}

