/* Svgalib Web Browser by Clocksoft */
/* (c) 1999 Clocksoft */
/* Main part of the browser wb0 with main() function*/

#include <vga.h>
#include <vgagl.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <math.h>

#include "gvars.h"
#include "line_breaker.h"
#include "scroller.h"
#include "watcher.h"
#include "printer.h"

#undef DEBUG_WB0

/*----------------------------------------------------------------------------*/

static int vga_mode=321; /* 320x200x256 */

/*----------------------------------------------------------------------------*/

int init_screen()
{
 if (scroller(SC_SETUP_VGA_MODE,&vga_mode))
 { 
  fprintf(stderr,"Unable to init vga mode.\n");
  return 1;
 }
 return 0;
}

void shutdown_screen()
{vga_setmode(TEXT);
}

int
interpret_args(int argc, char **argv)
{int a;
 unsigned char *txt;

 for (a=1;a<argc-1;a++)
 {txt=argv[a];
  switch(*txt++)
  {case 'm':
   {int mode=strtoul(txt,NULL,0);
    vga_mode=mode;
   }
   break;
   
   case 'h':
   {int height=strtoul(txt,NULL,0);
    if (height<=0) break;
    required_font_height=height;
   }
   break;
   
   case 'c':
   calibrate=1;
   break;
   
   case 'g':
   image_gamma=strtod(txt,NULL);
   break;
  }
 }
 if (argc>=2)
 {
  if (copy_string(&filename,argv[argc-1]))
  {
   return 1;
  }
 }
 return 0;
}

void deblank(unsigned char **p)
/* Shifts *p as much as is possible to skip all trailing blank chars. */
{unsigned char *t=*p;
 while(*t==' '||*t==9)t++;
 *p=t;
}

void add_into_enc_list(struct enc_member *m)
/* Inserts at the beginning. */
/* Always overwrites m->next */
{if (!enc_list)
 {/* Empty list */
  m->next=NULL; 
  enc_list=m;
 }
 else
 {m->next=enc_list;
  enc_list=m;
 }
}

#ifndef TEST
#define TEST(t) if(!strncmp(b,t,strlen(t))&&(txt=b+strlen(t),deblank(&txt),1))
int scan_config_line(unsigned char *b)
/* b is a string containing the line. */
{unsigned char *txt; /* After testing for token, points behind the token */
 unsigned char name[257],fname[257];
 unsigned char *name_ptr,*fname_ptr;
 struct enc_member *m;
 int param;
 
 if (!b||!*b) return 0;
 TEST("video_mode")
 {param=strtoul(txt,NULL,0);
  vga_mode=param;
  return 0;
 }
 TEST("font_height")
 {param=strtoul(txt,NULL,0);
  if (param<1) return 0;
  required_font_height=param;
  return 0;
 }
 TEST("path")
 {
  if (copy_string(&path,txt))
  {
   fprintf(stderr,"Unable to copy_string path.\n");
  }
  #ifdef DEBUG_WB0
   printf("path=%s.\n",path);
  #endif
  return 0;
 }
 TEST("encoding")
 {if (2!=sscanf(txt,"%256s%256s",name,fname)) return 0;
  if (!(m=(struct enc_member *)malloc(sizeof(struct enc_member)))||
   !(name_ptr=malloc(strlen(name)))||
   !(fname_ptr=malloc(strlen(fname))))
  {fprintf(stderr,"Unable to malloc small buffer to store encoding "
    "information.\n");
   return 1;
  }
  strcpy(name_ptr,name);
  strcpy(fname_ptr,fname);
  m->name=name_ptr;
  m->fname=fname_ptr;
  add_into_enc_list(m);
 }
 TEST("hquality")
 {
  hquality=strtoul(txt,NULL,0);
  return 0;
 }
 TEST("vquality")
 {
  vquality=strtoul(txt,NULL,0);
  return 0;
 }
 TEST("gamma_r") {monitor_gamma_r=strtod(txt,NULL);return 0;}
 TEST("gamma_g") {monitor_gamma_g=strtod(txt,NULL);return 0;}
 TEST("gamma_b") {monitor_gamma_b=strtod(txt,NULL);return 0;}
 TEST("required_gamma") {required_gamma=strtod(txt,NULL);return 0;}
 TEST("mul_r") {monitor_mul_r=strtod(txt,NULL);return 0;}
 TEST("mul_g") {monitor_mul_g=strtod(txt,NULL);return 0;}
 TEST("mul_b") {monitor_mul_b=strtod(txt,NULL);return 0;}
 TEST("shift_r") {monitor_shift_r=strtod(txt,NULL);return 0;}
 TEST("shift_g") {monitor_shift_g=strtod(txt,NULL);return 0;}
 TEST("shift_b") {monitor_shift_b=strtod(txt,NULL);return 0;}
 return 0;
}
#undef TEST
#else
#error Conflicting #define for TEST with something more essential
#endif

int read_config_file(unsigned char *fname)
{FILE *f;
 int len;
 unsigned char *buffer;
 const int buffer_length=1025; /* Don't forget the terminator 0 :-) */
 
 f=fopen(fname,"r");
 if (!f)
 {fprintf(stderr,"Unable to open config file %s.\n",fname);return 1;}
 buffer=malloc(buffer_length); 
 if (!buffer)
 {fclose(f);
  fprintf(stderr,"Unable to malloc %i bytes.\n",buffer_length);
  return 1;
 }
 scan_config_line("path /usr/local/src/wb0/font/bit/");
 while(!feof(f))
 {if (!fgets(buffer,buffer_length,f)) break;
  if (*buffer&&(len=strlen(buffer))&&buffer[len-1=='\n'])buffer[len-1]=0;
  if (scan_config_line(buffer))
  {fprintf(stderr,"Unable to scan config file line.\n");
   return 1;
  }
 }
 free(buffer);
 fclose(f);
 if (!enc_list)
 {fprintf(stderr,"No encodings specified.\n");
  return 1;
 }
 else
 {actual_encoding=enc_list; /* Take the first <=> the which appeared last in */
  /* the config file. */
 }
 return 0;
}

int
main(int argc,char **argv)
{
 if (vga_init())
 {
  fprintf(stderr,"vga_init() failed.\n");
  return 1;
 }
 if (copy_string(&filename,"")||copy_string(&path,""))
 {
  return 1;
 }
 if (read_config_file("/etc/wb0/wb0.config"))
 {fprintf(stderr,"Unable to read configuration file.\n");return 1;}
 if (interpret_args(argc,argv))
 {
  fprintf(stderr,"Unable to interpret args.\n");
  return 1;
 }
#ifdef DEBUG_WB0
 printf("main: path=%s.\n",path);
#endif
 if (init_screen()) goto err;
 watcher();
 shutdown_screen();
 fflush(stderr);
 return 0;
  
 err: 
 shutdown_screen();
 fflush(stderr);
 return 1;
}
